#!/usr/bin/env python
"""
Writer for Clustal format.
"""
from copy import copy

from cogent3.core.alignment import SequenceCollection


__author__ = "Jeremy Widmann"
__copyright__ = "Copyright 2007-2022, The Cogent Project"
__credits__ = ["Jeremy Widmann"]
__license__ = "BSD-3"
__version__ = "2023.2.12a1"
__maintainer__ = "Jeremy Widmann"
__email__ = "jeremy.widmann@colorado.edu"
__status__ = "Development"


def clustal_from_alignment(aln, wrap=None):
    """
    Parameters
    ----------
    aln
        can be an Alignment object or a dict
    wrap
        sequence line width.  Only available if sequences are
        aligned.

    Returns
    -------
    Returns a string in Clustal format
    """
    if not aln:
        return ""

    # get seq output order
    try:
        order = aln.RowOrder
    except:
        order = list(aln.keys())
        order.sort()

    seqs = SequenceCollection(aln)
    clustal_list = ["CLUSTAL\n"]

    if seqs.is_ragged():
        raise ValueError(
            "Sequences in alignment are not all the same length."
            + "Cannot generate Clustal format."
        )

    aln_len = seqs.seq_len
    # Get all labels
    labels = copy(seqs.names)

    # Find all label lengths in order to get padding.
    label_lengths = [len(l) for l in labels]
    label_max = max(label_lengths)
    max_spaces = label_max + 4

    # Get ordered seqs
    ordered_seqs = [seqs.named_seqs[label] for label in order]

    if wrap is not None:
        curr_ix = 0
        while curr_ix < aln_len:
            clustal_list.extend(
                [
                    f"{x}{' ' * (max_spaces - len(x))}{y[curr_ix:curr_ix + wrap]}"
                    for x, y in zip(order, ordered_seqs)
                ]
            )
            clustal_list.append("")
            curr_ix += wrap
    else:
        clustal_list.extend(
            [
                f"{x}{' ' * (max_spaces - len(x))}{y}"
                for x, y in zip(order, ordered_seqs)
            ]
        )
        clustal_list.append("")

    return "\n".join(clustal_list)



#!/usr/bin/env python
"""Unit tests for GFF and related parsers.
"""
import os

from io import StringIO
from pathlib import Path
from unittest import TestCase, main

from cogent3.parse.gff import *


__author__ = "Matthew Wakefield"
__copyright__ = "Copyright 2007-2022, The Cogent Project"
__credits__ = ["Matthew Wakefield"]
__license__ = "BSD-3"
__version__ = "2023.2.12a1"
__maintainer__ = "Matthew Wakefield"
__email__ = "wakefield@wehi.edu.au"
__status__ = "Production"

headers = [
    """##gff-version 2 
##source-version <source> <version text> 
##date <date> 
##Type <type> [<seqname>] 
##DNA <seqname>
##acggctcggattggcgctggatgatagatcagacgac
##...
##end-DNA
""",
    """##gff-version 2
""",
    "",
]

#    '<seqname>\t<source>\t<feature>\t<start>\t<end>\t<score>\t<strand>\t<frame>\t[attribute]\n'

data_lines = [
    (
        'seq1\tBLASTX\tsimilarity\t101\t235\t87.1\t+\t0\tTarget "HBA_HUMAN" 11 55 ; E_value 0.0003\n',
        (
            "seq1",
            "BLASTX",
            "similarity",
            100,
            235,
            "87.1",
            "+",
            "0",
            'Target "HBA_HUMAN" 11 55 ; E_value 0.0003',
            None,
        ),
    ),
    (
        'dJ102G20\tGD_mRNA\tcoding_exon\t7105\t7201\t.\t-\t2\tSequence "dJ102G20.C1.1"\n',
        (
            "dJ102G20",
            "GD_mRNA",
            "coding_exon",
            7201,
            7104,
            ".",
            "-",
            "2",
            'Sequence "dJ102G20.C1.1"',
            None,
        ),
    ),
    (
        "dJ102G20\tGD_mRNA\tcoding_exon\t7105\t7201\t.\t-\t2\t\n",
        ("dJ102G20", "GD_mRNA", "coding_exon", 7201, 7104, ".", "-", "2", "", None),
    ),
    (
        '12345\tSource with spaces\tfeature with spaces\t-100\t3600000000\t1e-5\t-\t.\tSequence "BROADO5" ; Note "This is a \\t tab containing \\n multi line comment"\n',
        (
            "12345",
            "Source with spaces",
            "feature with spaces",
            3600000000,
            101,
            "1e-5",
            "-",
            ".",
            'Sequence "BROADO5" ; Note "This is a \\t tab containing \\n multi line comment"',
            None,
        ),
    ),
]


class GffTest(TestCase):
    """Setup data for all the GFF parsers."""

    def testGffParserData(self):
        """Test gff_parser with valid data lines"""
        for (line, canned_result) in data_lines:
            result = next(gff_parser(StringIO(line)))
            canned_result = list(canned_result)
            self.assertEqual(result.pop("Attributes")["Info"], canned_result.pop(8))
            self.assertEqual(set(result.values()), set(canned_result))

    def test_gff_parser_headers(self):
        """Test gff_parser with valid data headers"""
        data = "".join([x[0] for x in data_lines])
        for header in headers:
            result = list(gff_parser(StringIO(header + data)))
            lines = [(x[0], list(x[1])) for x in data_lines]
            self.assertEqual(
                [l.pop("Attributes")["Info"] for l in result],
                [x[1].pop(8) for x in lines],
            )
            self.assertEqual(
                [set(l.values()) for l in result], [set(x[1]) for x in lines]
            )

    def test_parse_attributes_gff2(self):
        """Test the parse_attributes_gff2 method"""
        self.assertEqual(
            [
                parse_attributes_gff2(x[1][8], (x[1][3], x[1][4]))["ID"]
                for x in data_lines
            ],
            ["HBA_HUMAN", "dJ102G20.C1.1", "", "BROADO5"],
        )

    def test_gff2_parser_string(self):
        """Test the gff_parser works with a string filepath"""
        filepath = os.path.join("data/gff2_test.gff")
        for i, result in enumerate(gff_parser(filepath)):
            line = list(data_lines[i][1])
            self.assertEqual(result.pop("Attributes")["Info"], line.pop(8))
            self.assertEqual(set(result.values()), set(line))

    def test_gff2_parser_path(self):
        """Test the gff_parser works with a pathlib.Path filepath"""
        filepath = Path("data/gff2_test.gff")
        for i, result in enumerate(gff_parser(filepath)):
            line = list(data_lines[i][1])
            self.assertEqual(result.pop("Attributes")["Info"], line.pop(8))
            self.assertEqual(set(result.values()), set(line))

    def test_gff3_parser(self):
        """Test the gff_parser works on a gff3 file"""
        gff3_path = os.path.join("data/c_elegans_WS199_shortened_gff.gff3")
        for i, result in enumerate(gff_parser(gff3_path)):
            self.assertEqual(len(result), 10)
        # 15 total lines, but 2 comments
        self.assertEqual(i + 1, 15 - 2)


if __name__ == "__main__":
    main()



"""Unit tests for statistical tests and utility functions.
"""

from unittest import TestCase, main

from numpy import (
    arange,
    array,
    asarray,
    concatenate,
    fill_diagonal,
    isfinite,
    isnan,
    logical_and,
    ones,
    reshape,
    testing,
    tril,
)

from cogent3.maths.stats.number import NumberCounter
from cogent3.maths.stats.test import (
    ALT_HIGH,
    ALT_LOW,
    ALT_TWO_SIDED,
    ANOVA_one_way,
    G_2_by_2,
    G_fit,
    G_ind,
    MonteCarloP,
    ZeroExpectedError,
    _flatten_lower_triangle,
    _get_alternate,
    _get_rank,
    _permute_observations,
    bayes_updates,
    combinations,
    correlation,
    correlation_matrix,
    correlation_test,
    distance_matrix_permutation_test,
    f_two_sample,
    f_value,
    fisher,
    get_ltm_cells,
    get_values_from_matrix,
    is_symmetric_and_hollow,
    kendall_correlation,
    likelihoods,
    mantel,
    mantel_test,
    mc_t_two_sample,
    multiple_comparisons,
    multiple_inverse,
    multiple_n,
    mw_boot,
    mw_test,
    pearson,
    pearson_correlation,
    permute_2d,
    posteriors,
    regress,
    regress_major,
    regress_origin,
    regress_R2,
    regress_residuals,
    reverse_tails,
    safe_sum_p_log_p,
    sign_test,
    spearman,
    std,
    stdev_from_mean,
    t_one_observation,
    t_one_sample,
    t_paired,
    t_two_sample,
    tail,
    z_test,
)


__author__ = "Rob Knight"
__copyright__ = "Copyright 2007-2022, The Cogent Project"
__credits__ = [
    "Rob Knight",
    "Catherine Lozupone",
    "Gavin Huttley",
    "Sandra Smit",
    "Daniel McDonald",
    "Jai Ram Rideout",
    "Michael Dwan",
]
__license__ = "BSD-3"
__version__ = "2023.2.12a1"
__maintainer__ = "Gavin Huttley"
__email__ = "Gavin.Huttley@anu.edu.au"
__status__ = "Production"

from numpy.testing import assert_allclose, assert_equal


def is_prob(value):
    """helper function to establish a 0 <= value <= 1"""
    value = asarray(value)
    return logical_and(value >= 0, value <= 1.0).all()


def similar_means(observed, expected, pvalue=0.01):
    """False if observed p is lower than pvalue"""

    observed, expected = asarray(observed), asarray(expected)

    t, p = t_two_sample(observed, expected)

    # handle case where all elements were the same
    if p is None or not isfinite(p):

        if not observed.shape:
            observed = observed.reshape((1,))

        if not expected.shape:
            expected = expected.reshape((1,))

        if observed[0] == expected[0]:
            return True

    return p > pvalue


class TestsHelper(TestCase):
    """Class with utility methods useful for other tests."""

    def setUp(self):
        """Sets up variables used in the tests."""
        # How many times a p-value should be tested to fall in a given range
        # before failing the test.
        self.p_val_tests = 10

    def assertCorrectPValue(
        self, exp_min, exp_max, fn, args=None, kwargs=None, p_val_idx=0
    ):
        """Tests that the stochastic p-value falls in the specified range.

        Performs the test self.p_val_tests times and fails if the observed
        p-value does not fall into the specified range at least once. Each
        p-value is also tested that it falls in the range 0.0 to 1.0.

        This method assumes that fn is callable, and will unpack and pass args
        and kwargs to fn if they are provided. It also assumes that fn returns
        a single value (the p-value to be tested) or a tuple of results (any
        length greater than or equal to 1), with the p-value at position
        p_val_idx.

        This is primarily used for testing the Mantel and correlation_test
        functions.
        """
        found_match = False
        for i in range(self.p_val_tests):
            if args is not None and kwargs is not None:
                obs = fn(*args, **kwargs)
            elif args is not None:
                obs = fn(*args)
            elif kwargs is not None:
                obs = fn(**kwargs)
            else:
                obs = fn()

            try:
                p_val = float(obs)
            except TypeError:
                p_val = obs[p_val_idx]

            self.assertTrue(is_prob(p_val))
            if p_val >= exp_min and p_val <= exp_max:
                found_match = True
                break
        self.assertTrue(found_match)


class TestsTests(TestCase):
    """Tests miscellaneous functions."""

    def test_std(self):
        """Should produce a standard deviation of 1.0 for a std normal dist"""
        expected = 1.58113883008
        assert_allclose(std(array([1, 2, 3, 4, 5])), expected)

        expected_a = array([expected, expected, expected, expected, expected])
        a = array(
            [
                [1, 2, 3, 4, 5],
                [5, 1, 2, 3, 4],
                [4, 5, 1, 2, 3],
                [3, 4, 5, 1, 2],
                [2, 3, 4, 5, 1],
            ]
        )
        assert_allclose(std(a, axis=0), expected_a)
        assert_allclose(std(a, axis=1), expected_a)
        self.assertRaises(ValueError, std, a, 5)

    def test_std_2d(self):
        """Should produce from 2darray the same stdevs as scipy.stats.std"""
        inp = array([[1, 2, 3], [4, 5, 6]])
        exps = (  # tuple(scipy_std(inp, ax) for ax in [None, 0, 1])
            1.8708286933869707,
            array([2.12132034, 2.12132034, 2.12132034]),
            array([1.0, 1.0]),
        )
        results = tuple(std(inp, ax) for ax in [None, 0, 1])
        for obs, exp in zip(results, exps):
            testing.assert_almost_equal(obs, exp)

    def test_std_3d(self):
        """Should produce from 3darray the same std devs as scipy.stats.std"""
        inp3d = array([[[0, 2, 2], [3, 4, 5]], [[1, 9, 0], [9, 10, 1]]])  # 2,2,3
        exp3d = (  # for axis None, 0, 1, 2: calc from scipy.stats.std
            3.63901418552,
            array(
                [
                    [0.70710678, 4.94974747, 1.41421356],
                    [4.24264069, 4.24264069, 2.82842712],
                ]
            ),
            array(
                [
                    [2.12132034, 1.41421356, 2.12132034],
                    [5.65685425, 0.70710678, 0.70710678],
                ]
            ),
            array([[1.15470054, 1.0], [4.93288286, 4.93288286]]),
        )
        res = tuple(std(inp3d, ax) for ax in [None, 0, 1, 2])
        for obs, exp in zip(res, exp3d):
            testing.assert_almost_equal(obs, exp)

    def test_tail(self):
        """tail should return x/2 if test is true; 1-(x/2) otherwise"""
        assert_allclose(tail(0.25, "a" == "a"), 0.25 / 2)
        assert_allclose(tail(0.25, "a" != "a"), 1 - (0.25 / 2))

    def test_multiple_comparisons(self):
        """multiple_comparisons should match values from R"""
        assert_allclose(
            multiple_comparisons(1e-7, 10000), 1 - 0.9990005, rtol=1e-6, atol=1e-6
        )
        assert_allclose(multiple_comparisons(0.05, 10), 0.4012631)
        assert_allclose(multiple_comparisons(1e-20, 1), 1e-20)
        assert_allclose(multiple_comparisons(1e-300, 1), 1e-300)
        assert_allclose(multiple_comparisons(0.95, 3), 0.99987499999999996)
        assert_allclose(multiple_comparisons(0.75, 100), 0.999999999999679)
        assert_allclose(multiple_comparisons(0.5, 1000), 1)
        assert_allclose(multiple_comparisons(0.01, 1000), 0.99995682875259)
        assert_allclose(multiple_comparisons(0.5, 5), 0.96875)
        assert_allclose(multiple_comparisons(1e-20, 10), 1e-19)

    def test_multiple_inverse(self):
        """multiple_inverse should invert multiple_comparisons results"""
        # NOTE: multiple_inverse not very accurate close to 1
        assert_allclose(
            multiple_inverse(1 - 0.9990005, 10000), 1e-7, rtol=1e-6, atol=1e-6
        )
        assert_allclose(multiple_inverse(0.4012631, 10), 0.05, rtol=1e-6, atol=1e-6)
        assert_allclose(multiple_inverse(1e-20, 1), 1e-20)
        assert_allclose(multiple_inverse(1e-300, 1), 1e-300)
        assert_allclose(multiple_inverse(0.96875, 5), 0.5)
        assert_allclose(multiple_inverse(1e-19, 10), 1e-20)

    def test_multiple_n(self):
        """multiple_n should swap parameters in multiple_comparisons"""
        assert_allclose(multiple_n(1e-7, 1 - 0.9990005), 10000, rtol=1e-6, atol=1e-6)
        assert_allclose(multiple_n(0.05, 0.4012631), 10, rtol=1e-6, atol=1e-6)
        assert_allclose(multiple_n(1e-20, 1e-20), 1)
        assert_allclose(multiple_n(1e-300, 1e-300), 1)
        assert_allclose(multiple_n(0.95, 0.99987499999999996), 3)
        assert_allclose(multiple_n(0.5, 0.96875), 5)
        assert_allclose(multiple_n(1e-20, 1e-19), 10)

    def test_fisher(self):
        """fisher results should match p 795 Sokal and Rohlf"""
        assert_allclose(
            fisher([0.073, 0.086, 0.10, 0.080, 0.060]),
            0.0045957946540917905,
            rtol=1e-6,
            atol=1e-6,
        )

    def test_regress(self):
        """regression slope, intercept should match p 459 Sokal and Rohlf"""
        x = [0, 12, 29.5, 43, 53, 62.5, 75.5, 85, 93]
        y = [8.98, 8.14, 6.67, 6.08, 5.90, 5.83, 4.68, 4.20, 3.72]
        assert_allclose(regress(x, y), (-0.05322, 8.7038), 0.001)
        # higher precision from OpenOffice
        assert_allclose(regress(x, y), (-0.05322215, 8.70402730))

        # add test to confirm no overflow error with large numbers
        x = [32119, 33831]
        y = [2.28, 2.43]
        exp = (8.761682243e-05, -5.341209112e-01)
        assert_allclose(regress(x, y), exp, 0.001)

    def test_regress_origin(self):
        """regression slope constrained through origin should match Excel"""
        x = array([1, 2, 3, 4])
        y = array([4, 2, 6, 8])
        assert_allclose(regress_origin(x, y), (1.9333333, 0))

        # add test to confirm no overflow error with large numbers
        x = [32119, 33831]
        y = [2.28, 2.43]
        exp = (7.1428649481939822e-05, 0)
        assert_allclose(regress_origin(x, y), exp, 0.001)

    def test_regress_R2(self):
        """regress_R2 returns the R^2 value of a regression"""
        x = [1.0, 2.0, 3.0, 4.0, 5.0]
        y = [2.1, 4.2, 5.9, 8.4, 9.6]
        result = regress_R2(x, y)
        assert_allclose(result, 0.99171419347896)

    def test_regress_residuals(self):
        """regress_residuals reprts error for points in linear regression"""
        x = [1.0, 2.0, 3.0, 4.0, 5.0]
        y = [2.1, 4.2, 5.9, 8.4, 9.6]
        result = regress_residuals(x, y)
        assert_allclose(result, [-0.1, 0.08, -0.14, 0.44, -0.28])

    def test_stdev_from_mean(self):
        """stdev_from_mean returns num std devs from mean for each val in x"""
        x = [2.1, 4.2, 5.9, 8.4, 9.6]
        result = stdev_from_mean(x)
        assert_allclose(
            result,
            [
                -1.292463399014413,
                -0.60358696806764478,
                -0.045925095396451399,
                0.77416589382589174,
                1.1678095686526162,
            ],
        )

    def test_regress_major(self):
        """major axis regression should match p 589 Sokal and Rohlf"""
        # Note that the Sokal and Rohlf example flips the axes, such that the
        # equation is for explaining x in terms of y, not y in terms of x.
        # Behavior here is the reverse, for easy comparison with regress.
        y = [159, 179, 100, 45, 384, 230, 100, 320, 80, 220, 320, 210]
        x = [
            14.40,
            15.20,
            11.30,
            2.50,
            22.70,
            14.90,
            1.41,
            15.81,
            4.19,
            15.39,
            17.25,
            9.52,
        ]
        assert_allclose(regress_major(x, y), (18.93633, -32.55208), rtol=1e-6)

    def test_sign_test(self):
        """sign_test, should match values from R"""
        import numpy

        v = [
            ("two sided", 26, 50, 0.88772482734078251),
            ("less", 26, 50, 0.6641),
            ("l", 10, 50, 1.193066583837777e-05),
            ("hi", 30, 50, 0.1013193755322703),
            ("h", 0, 50, 1.0),
            ("2", 30, 50, 0.20263875106454063),
            ("h", 49, 50, 4.5297099404706387e-14),
            ("h", 50, 50, 8.8817841970012543e-16),
            ("2", numpy.int64(95), 124, 2.204644901720111e-09),
        ]
        for alt, success, trials, p in v:
            result = sign_test(success, trials, alt=alt)
            assert_allclose(result, p, rtol=1e-5)

    def test_permute_2d(self):
        """permute_2d permutes rows and cols of a matrix."""
        a = reshape(arange(9), (3, 3))
        assert_equal(permute_2d(a, [0, 1, 2]), a)
        assert_equal(permute_2d(a, [2, 1, 0]), array([[8, 7, 6], [5, 4, 3], [2, 1, 0]]))
        assert_equal(permute_2d(a, [1, 2, 0]), array([[4, 5, 3], [7, 8, 6], [1, 2, 0]]))

    def test_get_alternate(self):
        """correctly identifies the specified alternate hypothesis"""
        alt = _get_alternate("lo")
        self.assertEqual(alt, ALT_LOW)
        alt = _get_alternate("hi")
        self.assertEqual(alt, ALT_HIGH)
        alt = _get_alternate("2")
        self.assertEqual(alt, ALT_TWO_SIDED)
        with self.assertRaises(ValueError):
            _get_alternate("22")


class GTests(TestCase):
    """Tests implementation of the G tests for fit and independence."""

    def test_G_2_by_2_2tailed_equal(self):
        """G_2_by_2 should return 0 if all cell counts are equal"""
        assert_allclose(0, G_2_by_2(1, 1, 1, 1, False, False)[0], atol=1e-12)
        assert_allclose(0, G_2_by_2(100, 100, 100, 100, False, False)[0], atol=1e-12)
        assert_allclose(0, G_2_by_2(100, 100, 100, 100, True, False)[0], atol=1e-12)

    def test_G_2_by_2_bad_data(self):
        """G_2_by_2 should raise ValueError if any counts are negative"""
        self.assertRaises(ValueError, G_2_by_2, 1, -1, 1, 1)

    def test_G_2_by_2_2tailed_examples(self):
        """G_2_by_2 values should match examples in Sokal & Rohlf"""
        # example from p 731, Sokal and Rohlf (1995)
        # without correction
        assert_allclose(G_2_by_2(12, 22, 16, 50, False, False)[0], 1.33249, 0.0001)
        assert_allclose(G_2_by_2(12, 22, 16, 50, False, False)[1], 0.24836, 0.0001)
        # with correction
        assert_allclose(G_2_by_2(12, 22, 16, 50, True, False)[0], 1.30277, 0.0001)
        assert_allclose(G_2_by_2(12, 22, 16, 50, True, False)[1], 0.25371, 0.0001)

    def test_G_2_by_2_1tailed_examples(self):
        """G_2_by_2 values should match values from codon_binding program"""
        # first up...the famous arginine case
        assert_allclose(
            G_2_by_2(36, 16, 38, 106), (29.111609, 0), rtol=0.00001, atol=1e-6
        )
        # then some other miscellaneous positive and negative values
        assert_allclose(
            G_2_by_2(0, 52, 12, 132), (-7.259930, 0.996474), rtol=0.00001, atol=1e-6
        )
        assert_allclose(
            G_2_by_2(5, 47, 14, 130), (-0.000481, 0.508751), rtol=0.00001, atol=1e-6
        )
        assert_allclose(
            G_2_by_2(5, 47, 36, 108), (-6.065167, 0.993106), rtol=0.00001, atol=1e-6
        )

    def test_Gfit_unequal_lists(self):
        """Gfit should raise errors if lists unequal"""
        # lists must be equal
        self.assertRaises(ValueError, G_fit, [1, 2, 3], [1, 2])

    def test_Gfit_negative_observeds(self):
        """Gfit should raise ValueError if any observeds are negative."""
        self.assertRaises(ValueError, G_fit, [-1, 2, 3], [1, 2, 3])

    def test_Gfit_nonpositive_expecteds(self):
        """Gfit should raise ZeroExpectedError if expecteds are zero/negative"""
        self.assertRaises(ZeroExpectedError, G_fit, [1, 2, 3], [0, 1, 2])
        self.assertRaises(ZeroExpectedError, G_fit, [1, 2, 3], [-1, 1, 2])

    def test_Gfit_good_data(self):
        """Gfit tests for fit should match examples in Sokal and Rohlf"""
        # example from p. 699, Sokal and Rohlf (1995)
        obs = [63, 31, 28, 12, 39, 16, 40, 12]
        exp = [
            67.78125,
            22.59375,
            22.59375,
            7.53125,
            45.18750,
            15.06250,
            45.18750,
            15.06250,
        ]
        # without correction
        assert_allclose(G_fit(obs, exp, False)[0], 8.82397, 0.00002)
        assert_allclose(G_fit(obs, exp, False)[1], 0.26554, 0.00002)
        # with correction
        assert_allclose(G_fit(obs, exp)[0], 8.76938, 0.00002)
        assert_allclose(G_fit(obs, exp)[1], 0.26964, 0.00002)

        # example from p. 700, Sokal and Rohlf (1995)
        obs = [130, 46]
        exp = [132, 44]
        # without correction
        assert_allclose(G_fit(obs, exp, False)[0], 0.12002, 0.00002)
        assert_allclose(G_fit(obs, exp, False)[1], 0.72901, 0.00002)
        # with correction
        assert_allclose(G_fit(obs, exp)[0], 0.11968, 0.00002)
        assert_allclose(G_fit(obs, exp)[1], 0.72938, 0.00002)

    def test_safe_sum_p_log_p(self):
        """safe_sum_p_log_p should ignore zero elements, not raise error"""
        m = array([2, 4, 0, 8])
        self.assertEqual(safe_sum_p_log_p(m, 2), 2 * 1 + 4 * 2 + 8 * 3)

    def test_G_ind(self):
        """G test for independence should match Sokal and Rohlf p 738 values"""
        a = array([[29, 11], [273, 191], [8, 31], [64, 64]])
        assert_allclose(G_ind(a)[0], 28.59642)
        assert_allclose(G_ind(a, True)[0], 28.31244, rtol=1e-5)


class LikelihoodTests(TestCase):
    """Tests implementations of likelihood calculations."""

    def test_likelihoods_unequal_list_lengths(self):
        """likelihoods should raise ValueError if input lists unequal length"""
        self.assertRaises(ValueError, likelihoods, [1, 2], [1])

    def test_likelihoods_equal_priors(self):
        """likelihoods should equal Pr(D|H) if priors the same"""
        equal = [0.25, 0.25, 0.25, 0.25]
        unequal = [0.5, 0.25, 0.125, 0.125]
        equal_answer = [1, 1, 1, 1]
        unequal_answer = [2, 1, 0.5, 0.5]
        for obs, exp in zip(likelihoods(equal, equal), equal_answer):
            assert_allclose(obs, exp)

        for obs, exp in zip(likelihoods(unequal, equal), unequal_answer):
            assert_allclose(obs, exp)

    def test_likelihoods_equal_evidence(self):
        """likelihoods should return vector of 1's if evidence equal for all"""
        equal = [0.25, 0.25, 0.25, 0.25]
        unequal = [0.5, 0.25, 0.125, 0.125]
        equal_answer = [1, 1, 1, 1]
        not_unity = [0.7, 0.7, 0.7, 0.7]

        for obs, exp in zip(likelihoods(equal, unequal), equal_answer):
            assert_allclose(obs, exp)

        # should be the same if evidences don't sum to 1
        for obs, exp in zip(likelihoods(not_unity, unequal), equal_answer):
            assert_allclose(obs, exp)

    def test_likelihoods_unequal_evidence(self):
        """likelihoods should update based on weighted sum if evidence unequal"""
        not_unity = [1, 0.5, 0.25, 0.25]
        unequal = [0.5, 0.25, 0.125, 0.125]
        products = [1.4545455, 0.7272727, 0.3636364, 0.3636364]

        # if priors and evidence both unequal, likelihoods should change
        # (calculated using StarCalc)
        for obs, exp in zip(likelihoods(not_unity, unequal), products):
            assert_allclose(obs, exp)

    def test_posteriors_unequal_lists(self):
        """posteriors should raise ValueError if input lists unequal lengths"""
        self.assertRaises(ValueError, posteriors, [1, 2, 3], [1])

    def test_posteriors_good_data(self):
        """posteriors should return products of paired list elements"""
        first = [0, 0.25, 0.5, 1, 0.25]
        second = [0.25, 0.5, 0, 0.1, 1]
        product = [0, 0.125, 0, 0.1, 0.25]
        for obs, exp in zip(posteriors(first, second), product):
            assert_allclose(obs, exp)


class BayesUpdateTests(TestCase):
    """Tests implementation of Bayes calculations"""

    def setUp(self):
        first = [0.25, 0.25, 0.25]
        second = [0.1, 0.75, 0.3]
        third = [0.95, 1e-10, 0.2]
        fourth = [0.01, 0.9, 0.1]
        bad = [1, 2, 1, 1, 1]
        self.bad = [first, bad, second, third]
        self.test = [first, second, third, fourth]
        self.permuted = [fourth, first, third, second]
        self.deleted = [second, fourth, third]
        self.extra = [first, second, first, third, first, fourth, first]

        # BEWARE: low precision in second item, so need to adjust threshold
        # for assertFloatEqual accordingly (and use assertFloatEqualAbs).
        self.result = [0.136690646154, 0.000000009712, 0.863309344133]

    def test_bayes_updates_bad_data(self):
        """bayes_updates should raise ValueError on unequal-length lists"""
        self.assertRaises(ValueError, bayes_updates, self.bad)

    def test_bayes_updates_good_data(self):
        """bayes_updates should match hand calculations of probability updates"""
        # result for first -> fourth calculated by hand
        for obs, exp in zip(bayes_updates(self.test), self.result):
            assert_allclose(obs, exp, rtol=1e-11, atol=1e-6)

    def test_bayes_updates_permuted(self):
        """bayes_updates should not be affected by order of inputs"""
        for obs, exp in zip(bayes_updates(self.permuted), self.result):
            assert_allclose(obs, exp, rtol=1e-11, atol=1e-6)

    def test_bayes_update_nondiscriminating(self):
        """bayes_updates should be unaffected by extra nondiscriminating data"""
        # deletion of non-discriminating evidence should not affect result
        for obs, exp in zip(bayes_updates(self.deleted), self.result):
            assert_allclose(obs, exp, rtol=1e-5, atol=1e-6)
        # additional non-discriminating evidence should not affect result
        for obs, exp in zip(bayes_updates(self.extra), self.result):
            assert_allclose(obs, exp, rtol=1e-5, atol=1e-6)


class StatTests(TestsHelper):
    """Tests that the t and z tests are implemented correctly"""

    def setUp(self):
        super(StatTests, self).setUp()

        self.x = [
            7.33,
            7.49,
            7.27,
            7.93,
            7.56,
            7.81,
            7.46,
            6.94,
            7.49,
            7.44,
            7.95,
            7.47,
            7.04,
            7.10,
            7.64,
        ]

        self.y = [
            7.53,
            7.70,
            7.46,
            8.21,
            7.81,
            8.01,
            7.72,
            7.13,
            7.68,
            7.66,
            8.11,
            7.66,
            7.20,
            7.25,
            7.79,
        ]

    def test_t_paired_2tailed(self):
        """t_paired should match values from Sokal & Rohlf p 353"""
        x, y = self.x, self.y
        # check value of t and the probability for 2-tailed
        assert_allclose(t_paired(y, x)[0], 19.7203, 1e-4)
        assert_allclose(t_paired(y, x)[1], 1.301439e-11, 1e-4)

    def test_t_paired_no_variance(self):
        """t_paired should return None if lists are invariant"""
        x = [1, 1, 1]
        y = [0, 0, 0]
        self.assertEqual(t_paired(x, x), (None, None))
        self.assertEqual(t_paired(x, y), (None, None))

    def test_t_paired_1tailed(self):
        """t_paired should match pre-calculated 1-tailed values"""
        x, y = self.x, self.y
        # check probability for 1-tailed low and high
        assert_allclose(t_paired(y, x, "low")[1], 1 - (1.301439e-11 / 2), 1e-4)
        assert_allclose(t_paired(x, y, "high")[1], 1 - (1.301439e-11 / 2), 1e-4)
        assert_allclose(t_paired(y, x, "high")[1], 1.301439e-11 / 2, 1e-4)
        assert_allclose(t_paired(x, y, "low")[1], 1.301439e-11 / 2, 1e-4)

    def test_t_paired_specific_difference(self):
        """t_paired should allow a specific difference to be passed"""
        x, y = self.x, self.y
        # difference is 0.2, so test should be non-significant if 0.2 passed
        self.assertFalse(t_paired(y, x, exp_diff=0.2)[0] > 1e-10)
        # same, except that reversing list order reverses sign of difference
        self.assertFalse(t_paired(x, y, exp_diff=-0.2)[0] > 1e-10)
        # check that there's no significant difference from the true mean
        assert_allclose(t_paired(y, x, exp_diff=0.2)[1], 1, 1e-4)

    def test_t_paired_bad_data(self):
        """t_paired should raise ValueError on lists of different lengths"""
        self.assertRaises(ValueError, t_paired, self.y, [1, 2, 3])

    def test_t_two_sample(self):
        """t_two_sample should match example on p.225 of Sokal and Rohlf"""
        I = array([7.2, 7.1, 9.1, 7.2, 7.3, 7.2, 7.5])
        II = array([8.8, 7.5, 7.7, 7.6, 7.4, 6.7, 7.2])
        assert_allclose(t_two_sample(I, II), (-0.1184, 0.45385 * 2), rtol=0.01)

    def test_t_two_sample_no_variance(self):
        """t_two_sample should properly handle lists that are invariant"""
        # By default should return (None, None) to mimic R's t.test.
        x = array([1, 1.0, 1])
        y = array([0, 0, 0.0])
        self.assertEqual(t_two_sample(x, x), (None, None))
        self.assertEqual(t_two_sample(x, y), (None, None))

        # Test none_on_zero_variance=False on various tail types. We use
        # self.assertEqual instead of assert_allclose because the latter
        # sees inf and -inf as being equal.

        # Two tailed: a < b
        self.assertEqual(
            t_two_sample(y, x, none_on_zero_variance=False), (float("-inf"), 0.0)
        )

        # Two tailed: a > b
        self.assertEqual(
            t_two_sample(x, y, none_on_zero_variance=False), (float("inf"), 0.0)
        )

        # One-tailed 'high': a < b
        self.assertEqual(
            t_two_sample(y, x, tails="high", none_on_zero_variance=False),
            (float("-inf"), 1.0),
        )

        # One-tailed 'high': a > b
        self.assertEqual(
            t_two_sample(x, y, tails="high", none_on_zero_variance=False),
            (float("inf"), 0.0),
        )

        # One-tailed 'low': a < b
        self.assertEqual(
            t_two_sample(y, x, tails="low", none_on_zero_variance=False),
            (float("-inf"), 0.0),
        )

        # One-tailed 'low': a > b
        self.assertEqual(
            t_two_sample(x, y, tails="low", none_on_zero_variance=False),
            (float("inf"), 1.0),
        )

        # Should still receive (None, None) if the lists have no variance and
        # have the same single value.
        self.assertEqual(t_two_sample(x, x, none_on_zero_variance=False), (None, None))
        self.assertEqual(
            t_two_sample(x, [1, 1], none_on_zero_variance=False), (None, None)
        )

    def test_t_two_sample_invalid_input(self):
        """t_two_sample should raise an error on invalid input."""
        self.assertRaises(ValueError, t_two_sample, [1, 2, 3], [4, 5, 6], tails="foo")

    def test_t_one_sample(self):
        """t_one_sample results should match those from R"""
        x = array(list(range(-5, 5)))
        y = array(list(range(-1, 10)))
        assert_allclose(t_one_sample(x), (-0.5222, 0.6141), 1e-4)
        assert_allclose(t_one_sample(y), (4, 0.002518), rtol=1e-3)
        # do some one-tailed tests as well
        assert_allclose(t_one_sample(y, tails="low"), (4, 0.9987), rtol=1e-3)
        assert_allclose(t_one_sample(y, tails="high"), (4, 0.001259), rtol=1e-3)

    def test_t_two_sample_switch(self):
        """t_two_sample should call t_one_observation if 1 item in sample."""
        sample = array([4.02, 3.88, 3.34, 3.87, 3.18])
        x = array([3.02])
        assert_allclose(t_two_sample(x, sample), (-1.5637254, 0.1929248))
        assert_allclose(t_two_sample(sample, x), (1.5637254, 0.1929248))

        # can't do the test if both samples have single item
        self.assertTrue(isnan(t_two_sample(x, x)).all())

        # Test special case if t=0.
        assert_allclose(t_two_sample([2], [1, 2, 3]), (0.0, 1.0))
        assert_allclose(t_two_sample([1, 2, 3], [2]), (0.0, 1.0))

    def test_t_one_observation(self):
        """t_one_observation should match p. 228 of Sokal and Rohlf"""
        sample = array([4.02, 3.88, 3.34, 3.87, 3.18])
        x = 3.02
        # note that this differs after the 3rd decimal place from what's in the
        # book, because Sokal and Rohlf round their intermediate steps...
        assert_allclose(t_one_observation(x, sample), (-1.5637254, 0.1929248))

    def test_t_one_observation_no_variance(self):
        """t_one_observation should correctly handle an invariant list."""
        sample = array([1.0, 1.0, 1.0])

        # Can't perform test if invariant list's single value matches x,
        # regardless of none_on_zero_variance.
        self.assertEqual(t_one_observation(1, sample), (None, None))
        self.assertEqual(
            t_one_observation(1, sample, none_on_zero_variance=False), (None, None)
        )

        # Test correct handling of none_on_zero_variance.
        self.assertEqual(t_one_observation(2, sample), (None, None))
        self.assertEqual(
            t_one_observation(2, sample, none_on_zero_variance=False),
            (float("inf"), 0.0),
        )
        self.assertEqual(
            t_one_observation(2, sample, none_on_zero_variance=False, tails="low"),
            (float("inf"), 1.0),
        )

    def test_mc_t_two_sample(self):
        """Test gives correct results with valid input data."""
        # Verified against R's t.test() and Deducer::perm.t.test().

        # With numpy array as input.
        exp = (-0.11858541225631833, 0.90756579317867436)
        I = array([7.2, 7.1, 9.1, 7.2, 7.3, 7.2, 7.5])
        II = array([8.8, 7.5, 7.7, 7.6, 7.4, 6.7, 7.2])
        obs = mc_t_two_sample(I, II)
        assert_allclose(obs[:2], exp)
        self.assertEqual(len(obs[2]), 999)
        self.assertCorrectPValue(0.8, 0.9, mc_t_two_sample, [I, II], p_val_idx=3)

        # With python list as input.
        exp = (-0.11858541225631833, 0.90756579317867436)
        I = [7.2, 7.1, 9.1, 7.2, 7.3, 7.2, 7.5]
        II = [8.8, 7.5, 7.7, 7.6, 7.4, 6.7, 7.2]
        obs = mc_t_two_sample(I, II)
        assert_allclose(obs[:2], exp)
        self.assertEqual(len(obs[2]), 999)
        self.assertCorrectPValue(0.8, 0.9, mc_t_two_sample, [I, II], p_val_idx=3)

        exp = (-0.11858541225631833, 0.45378289658933718)
        obs = mc_t_two_sample(I, II, tails="low")
        assert_allclose(obs[:2], exp)
        self.assertEqual(len(obs[2]), 999)
        self.assertCorrectPValue(
            0.4, 0.47, mc_t_two_sample, [I, II], {"tails": "low"}, p_val_idx=3
        )

        exp = (-0.11858541225631833, 0.54621710341066287)
        obs = mc_t_two_sample(I, II, tails="high", permutations=99)
        assert_allclose(obs[:2], exp)
        self.assertEqual(len(obs[2]), 99)
        self.assertCorrectPValue(
            0.4,
            0.62,
            mc_t_two_sample,
            [I, II],
            {"tails": "high", "permutations": 99},
            p_val_idx=3,
        )

        exp = (-2.8855783649036986, 0.99315596652421401)
        obs = mc_t_two_sample(I, II, tails="high", permutations=99, exp_diff=1)
        assert_allclose(obs[:2], exp)
        self.assertEqual(len(obs[2]), 99)
        self.assertCorrectPValue(
            0.55,
            0.99,
            mc_t_two_sample,
            [I, II],
            {"tails": "high", "permutations": 99, "exp_diff": 1},
            p_val_idx=3,
        )

    def test_mc_t_two_sample_unbalanced_obs(self):
        """Test gives correct results with unequal number of obs per sample."""
        # Verified against R's t.test() and Deducer::perm.t.test().
        exp = (-0.10302479888889175, 0.91979753020527177)
        I = array([7.2, 7.1, 9.1, 7.2, 7.3, 7.2])
        II = array([8.8, 7.5, 7.7, 7.6, 7.4, 6.7, 7.2])
        obs = mc_t_two_sample(I, II)
        assert_allclose(obs[:2], exp)
        self.assertEqual(len(obs[2]), 999)
        self.assertCorrectPValue(0.8, 0.9, mc_t_two_sample, [I, II], p_val_idx=3)

    def test_mc_t_two_sample_single_obs_sample(self):
        """Test works correctly with one sample having a single observation."""
        sample = array([4.02, 3.88, 3.34, 3.87, 3.18])
        x = array([3.02])
        exp = (-1.5637254, 0.1929248)
        obs = mc_t_two_sample(x, sample)
        assert_allclose(obs[:2], exp)
        assert_allclose(len(obs[2]), 999)
        self.assertTrue(is_prob(obs[3]))

        exp = (1.5637254, 0.1929248)
        obs = mc_t_two_sample(sample, x)
        assert_allclose(obs[:2], exp)
        assert_allclose(len(obs[2]), 999)
        self.assertTrue(is_prob(obs[3]))

        # Test the case where we can have no variance in the permuted lists.
        x = array([1, 1, 2])
        y = array([1])
        exp = (0.5, 0.666666666667)
        obs = mc_t_two_sample(x, y)
        assert_allclose(obs[:2], exp)
        assert_allclose(len(obs[2]), 999)
        self.assertTrue(is_prob(obs[3]))

    def test_mc_t_two_sample_no_perms(self):
        """Test gives empty permutation results if no perms are given."""
        exp = (-0.11858541225631833, 0.90756579317867436, [], None)
        I = array([7.2, 7.1, 9.1, 7.2, 7.3, 7.2, 7.5])
        II = array([8.8, 7.5, 7.7, 7.6, 7.4, 6.7, 7.2])
        obs = mc_t_two_sample(I, II, permutations=0)
        assert_allclose(obs[:2], exp[:2])
        assert_equal(obs[2:], exp[2:])

    def test_mc_t_two_sample_no_mc(self):
        """Test no MC stats if initial t-test is bad."""
        x = array([1, 1, 1])
        self.assertEqual(mc_t_two_sample(x, x), (None, None, [], None))

    def test_mc_t_two_sample_no_variance(self):
        """Test input with no variance. Should match Deducer::perm.t.test."""
        x = array([1, 1, 1])
        y = array([2, 2, 2])

        exp = (float("-inf"), 0.0)
        obs = mc_t_two_sample(x, y, permutations=1000)

        self.assertEqual(obs[:2], exp)
        self.assertEqual(len(obs[2]), 1000)
        self.assertCorrectPValue(
            0.09, 0.11, mc_t_two_sample, [x, y], {"permutations": 1000}, p_val_idx=3
        )

        exp = (float("inf"), 0.0)
        obs = mc_t_two_sample(y, x, permutations=1000)

        self.assertEqual(obs[:2], exp)
        self.assertEqual(len(obs[2]), 1000)
        self.assertCorrectPValue(
            0.09, 0.11, mc_t_two_sample, [y, x], {"permutations": 1000}, p_val_idx=3
        )

        exp = (float("-inf"), 1.0)
        obs = mc_t_two_sample(x, y, permutations=1000, tails="high")

        self.assertEqual(obs[:2], exp)
        self.assertEqual(len(obs[2]), 1000)
        self.assertCorrectPValue(
            0.9999,
            1.0,
            mc_t_two_sample,
            [x, y],
            {"permutations": 1000, "tails": "high"},
            p_val_idx=3,
        )

        exp = (float("-inf"), 0.0)
        obs = mc_t_two_sample(x, y, permutations=1000, tails="low")

        self.assertEqual(obs[:2], exp)
        self.assertEqual(len(obs[2]), 1000)
        self.assertCorrectPValue(
            0.04,
            0.051,
            mc_t_two_sample,
            [x, y],
            {"permutations": 1000, "tails": "low"},
            p_val_idx=3,
        )

    def test_mc_t_two_sample_no_permuted_variance(self):
        """Test with chance of getting no variance with some perms."""
        # Verified against R's t.test() and Deducer::perm.t.test().
        x = array([1, 1, 2])
        y = array([2, 2, 1])

        exp = (-0.70710678118654791, 0.51851851851851838)
        obs = mc_t_two_sample(x, y, permutations=10000)

        assert_allclose(obs[:2], exp)
        self.assertEqual(len(obs[2]), 10000)
        self.assertCorrectPValue(
            0.97, 1.0, mc_t_two_sample, [x, y], {"permutations": 10000}, p_val_idx=3
        )

    def test_mc_t_two_sample_invalid_input(self):
        """Test fails on various invalid input."""
        self.assertRaises(
            ValueError, mc_t_two_sample, [1, 2, 3], [4.0, 5.0, 4.0], tails="foo"
        )
        self.assertRaises(
            ValueError, mc_t_two_sample, [1, 2, 3], [4.0, 5.0, 4.0], permutations=-1
        )
        self.assertRaises(ValueError, mc_t_two_sample, [1], [4.0])
        self.assertRaises(ValueError, mc_t_two_sample, [1, 2], [])

    def test_permute_observations(self):
        """Test works correctly on small input dataset."""
        I = [10, 20.0, 1]
        II = [2, 4, 5, 7]
        obs = _permute_observations(I, II, 1)
        self.assertEqual(len(obs[0]), 1)
        self.assertEqual(len(obs[1]), 1)
        self.assertEqual(len(obs[0][0]), len(I))
        self.assertEqual(len(obs[1][0]), len(II))
        assert_allclose(sorted(concatenate((obs[0][0], obs[1][0]))), sorted(I + II))

    def test_reverse_tails(self):
        """reverse_tails should return 'high' if tails was 'low' or vice versa"""
        self.assertEqual(reverse_tails("high"), "low")
        self.assertEqual(reverse_tails("low"), "high")
        self.assertEqual(reverse_tails(None), ALT_TWO_SIDED)

    def test_tail(self):
        """tail should return prob/2 if test is true, or 1-(prob/2) if false"""
        assert_allclose(tail(0.25, True), 0.125)
        assert_allclose(tail(0.25, False), 0.875)
        assert_allclose(tail(1, True), 0.5)
        assert_allclose(tail(1, False), 0.5)
        assert_allclose(tail(0, True), 0)
        assert_allclose(tail(0, False), 1)

    def test_z_test(self):
        """z_test should give correct values"""
        sample = array([1, 2, 3, 4, 5])
        assert_allclose(z_test(sample, 3, 1), (0, 1))
        assert_allclose(z_test(sample, 3, 2, "high"), (0, 0.5))
        assert_allclose(z_test(sample, 3, 2, "low"), (0, 0.5))
        # check that population mean and variance, and tails, can be set OK.
        assert_allclose(
            z_test(sample, 0, 1), (6.7082039324993694, 1.9703444711798951e-11)
        )
        assert_allclose(
            z_test(sample, 1, 10), (0.44721359549995793, 0.65472084601857694)
        )
        assert_allclose(
            z_test(sample, 1, 10, "high"),
            (0.44721359549995793, 0.65472084601857694 / 2),
        )
        assert_allclose(
            z_test(sample, 1, 10, "low"),
            (0.44721359549995793, 1 - (0.65472084601857694 / 2)),
        )


class CorrelationTests(TestsHelper):
    """Tests of correlation coefficients and Mantel test."""

    def setUp(self):
        """Sets up variables used in the tests."""
        super(CorrelationTests, self).setUp()

        # For testing spearman and correlation_test using method='spearman'.
        # Taken from the Spearman wikipedia article. Also used for testing
        # Pearson (verified with R).
        self.data1 = [106, 86, 100, 101, 99, 103, 97, 113, 112, 110]
        self.data2 = [7, 0, 27, 50, 28, 29, 20, 12, 6, 17]

        # For testing spearman.
        self.a = [1, 2, 4, 3, 1, 6, 7, 8, 10, 4]
        self.b = [2, 10, 20, 1, 3, 7, 5, 11, 6, 13]
        self.c = [7, 1, 20, 13, 3, 57, 5, 121, 2, 9]
        self.r = (1.7, 10, 20, 1.7, 3, 7, 5, 11, 6.5, 13)
        self.x = (1, 2, 4, 3, 1, 6, 7, 8, 10, 4, 100, 2, 3, 77)

        # Ranked copies for testing spearman.
        self.b_ranked = [2, 7, 10, 1, 3, 6, 4, 8, 5, 9]
        self.c_ranked = [5, 1, 8, 7, 3, 9, 4, 10, 2, 6]

    def test_mantel(self):
        """mantel should be significant for same matrix, not for random"""
        a = reshape(arange(25), (5, 5))
        a = tril(a) + tril(a).T
        fill_diagonal(a, 0)
        b = a.copy()
        # closely related -- should be significant
        self.assertCorrectPValue(0.0, 0.049, mantel, (a, b, 1000))

        c = reshape(ones(25), (5, 5))
        c[0, 1] = 3.0
        c[1, 0] = 3.0
        fill_diagonal(c, 0)
        # not related -- should not be significant
        self.assertCorrectPValue(0.06, 1.0, mantel, (a, c, 1000))

    def test_mantel_test_one_sided_greater(self):
        """Test one-sided mantel test (greater)."""
        # This test output was verified by R (their mantel function does a
        # one-sided greater test).
        m1 = array([[0, 1, 2], [1, 0, 3], [2, 3, 0]])
        m2 = array([[0, 2, 7], [2, 0, 6], [7, 6, 0]])
        p, stat, perms = mantel_test(m1, m1, 999, alt="greater")
        assert_allclose(stat, 1.0)
        self.assertEqual(len(perms), 999)
        self.assertCorrectPValue(
            0.09, 0.25, mantel_test, (m1, m1, 999), {"alt": "greater"}
        )

        p, stat, perms = mantel_test(m1, m2, 999, alt="greater")
        assert_allclose(stat, 0.755928946018)
        self.assertEqual(len(perms), 999)
        self.assertCorrectPValue(
            0.2, 0.5, mantel_test, (m1, m2, 999), {"alt": "greater"}
        )

    def test_mantel_test_one_sided_less(self):
        """Test one-sided mantel test (less)."""
        # This test output was verified by R (their mantel function does a
        # one-sided greater test, but I modified their output to do a one-sided
        # less test).
        m1 = array([[0, 1, 2], [1, 0, 3], [2, 3, 0]])
        m2 = array([[0, 2, 7], [2, 0, 6], [7, 6, 0]])
        m3 = array([[0, 0.5, 0.25], [0.5, 0, 0.1], [0.25, 0.1, 0]])
        p, stat, perms = mantel_test(m1, m1, 999, alt="less")
        assert_allclose(p, 1.0)
        assert_allclose(stat, 1.0)
        self.assertEqual(len(perms), 999)

        p, stat, perms = mantel_test(m1, m2, 999, alt="less")
        assert_allclose(stat, 0.755928946018)
        self.assertEqual(len(perms), 999)
        self.assertCorrectPValue(0.6, 1.0, mantel_test, (m1, m2, 999), {"alt": "less"})

        p, stat, perms = mantel_test(m1, m3, 999, alt="less")
        assert_allclose(stat, -0.989743318611)
        self.assertEqual(len(perms), 999)
        self.assertCorrectPValue(0.1, 0.25, mantel_test, (m1, m3, 999), {"alt": "less"})

    def test_mantel_test_two_sided(self):
        """Test two-sided mantel test."""
        # This test output was verified by R (their mantel function does a
        # one-sided greater test, but I modified their output to do a two-sided
        # test).
        m1 = array([[0, 1, 2], [1, 0, 3], [2, 3, 0]])
        m2 = array([[0, 2, 7], [2, 0, 6], [7, 6, 0]])
        m3 = array([[0, 0.5, 0.25], [0.5, 0, 0.1], [0.25, 0.1, 0]])
        p, stat, perms = mantel_test(m1, m1, 999, alt="two sided")
        assert_allclose(stat, 1.0)
        self.assertEqual(len(perms), 999)
        self.assertCorrectPValue(
            0.20, 0.45, mantel_test, (m1, m1, 999), {"alt": "two sided"}
        )

        p, stat, perms = mantel_test(m1, m2, 999, alt="two sided")
        assert_allclose(stat, 0.755928946018)
        self.assertEqual(len(perms), 999)
        self.assertCorrectPValue(
            0.6, 0.75, mantel_test, (m1, m2, 999), {"alt": "two sided"}
        )

        p, stat, perms = mantel_test(m1, m3, 999, alt="two sided")
        assert_allclose(stat, -0.989743318611)
        self.assertEqual(len(perms), 999)
        self.assertCorrectPValue(
            0.2, 0.45, mantel_test, (m1, m3, 999), {"alt": "two sided"}
        )

    def test_mantel_test_invalid_distance_matrix(self):
        """Test mantel test with invalid distance matrix."""
        # Single asymmetric, non-hollow distance matrix.
        self.assertRaises(
            ValueError,
            mantel_test,
            array([[1, 2], [3, 4]]),
            array([[0, 0], [0, 0]]),
            999,
        )

        # Two asymmetric distance matrices.
        self.assertRaises(
            ValueError,
            mantel_test,
            array([[0, 2], [3, 0]]),
            array([[0, 1], [0, 0]]),
            999,
        )

    def test_mantel_test_invalid_input(self):
        """Test mantel test with invalid input."""
        self.assertRaises(
            ValueError, mantel_test, array([[1]]), array([[1]]), 999, alt="foo"
        )
        self.assertRaises(
            ValueError, mantel_test, array([[1]]), array([[1, 2], [3, 4]]), 999
        )
        self.assertRaises(ValueError, mantel_test, array([[1]]), array([[1]]), 0)
        self.assertRaises(ValueError, mantel_test, array([[1]]), array([[1]]), -1)

    def test_is_symmetric_and_hollow(self):
        """Should correctly test for symmetry and hollowness of dist mats."""
        self.assertTrue(is_symmetric_and_hollow(array([[0, 1], [1, 0]])))
        self.assertTrue(is_symmetric_and_hollow(array([[0, 1], [1, 0]])))
        self.assertTrue(is_symmetric_and_hollow(array([[0.0, 0], [0.0, 0]])))
        self.assertTrue(not is_symmetric_and_hollow(array([[0.001, 1], [1, 0]])))
        self.assertTrue(not is_symmetric_and_hollow(array([[0, 1.1], [1, 0]])))
        self.assertTrue(not is_symmetric_and_hollow(array([[0.5, 1.1], [1, 0]])))

    def test_flatten_lower_triangle(self):
        """Test flattening various dms' lower triangulars."""
        self.assertEqual(_flatten_lower_triangle(array([[8]])), [])
        self.assertEqual(_flatten_lower_triangle(array([[1, 2], [3, 4]])), [3])
        self.assertEqual(
            _flatten_lower_triangle(array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])), [4, 7, 8]
        )

    def test_pearson(self):
        """Test pearson correlation method on valid data."""
        # This test output was verified by R.
        assert_allclose(pearson([1, 2], [1, 2]), 1.0)
        assert_allclose(pearson([1, 2, 3], [1, 2, 3]), 1.0)
        assert_allclose(pearson([1, 2, 3], [1, 2, 4]), 0.9819805)

    def test_pearson_invalid_input(self):
        """Test running pearson on bad input."""
        self.assertRaises(ValueError, pearson, [1.4, 2.5], [5.6, 8.8, 9.0])
        self.assertRaises(ValueError, pearson, [1.4], [5.6])

    def test_spearman(self):
        """Test the spearman function with valid input."""
        # One vector has no ties.
        exp = 0.3719581
        obs = spearman(self.a, self.b)
        assert_allclose(obs, exp)

        # Both vectors have no ties.
        exp = 0.2969697
        obs = spearman(self.b, self.c)
        assert_allclose(obs, exp)

        # Both vectors have ties.
        exp = 0.388381
        obs = spearman(self.a, self.r)
        assert_allclose(obs, exp)

        exp = -0.17575757575757578
        obs = spearman(self.data1, self.data2)
        assert_allclose(obs, exp)

    def test_spearman_no_variation(self):
        """Test the spearman function with a vector having no variation."""
        exp = 0.0
        obs = spearman([1, 1, 1], [1, 2, 3])
        assert_allclose(obs, exp)

    def test_spearman_ranked(self):
        """Test the spearman function with a vector that is already ranked."""
        exp = 0.2969697
        obs = spearman(self.b_ranked, self.c_ranked)
        assert_allclose(obs, exp)

    def test_spearman_one_obs(self):
        """Test running spearman on a single observation."""
        self.assertRaises(ValueError, spearman, [1.0], [5.0])

    def test_spearman_invalid_input(self):
        """Test the spearman function with invalid input."""
        self.assertRaises(ValueError, spearman, [], [])
        self.assertRaises(ValueError, spearman, self.a, [])
        self.assertRaises(TypeError, spearman, {0: 2}, [1, 2, 3])

    def test_get_rank(self):
        """Test the _get_rank function with valid input."""
        exp = (
            [1.5, 3.5, 7.5, 5.5, 1.5, 9.0, 10.0, 11.0, 12.0, 7.5, 14.0, 3.5, 5.5, 13.0],
            4,
        )
        obs = _get_rank(self.x)
        assert_allclose(obs[0], exp[0])
        self.assertEqual(obs[1], exp[1])

        exp = ([1.5, 3.0, 5.5, 4.0, 1.5, 7.0, 8.0, 9.0, 10.0, 5.5], 2)
        obs = _get_rank(self.a)
        assert_allclose(obs[0], exp[0])
        self.assertEqual(obs[1], exp[1])

        exp = ([2, 7, 10, 1, 3, 6, 4, 8, 5, 9], 0)
        obs = _get_rank(self.b)
        assert_allclose(obs[0], exp[0])
        self.assertEqual(obs[1], exp[1])

        exp = ([1.5, 7.0, 10.0, 1.5, 3.0, 6.0, 4.0, 8.0, 5.0, 9.0], 1)
        obs = _get_rank(self.r)
        assert_allclose(obs[0], exp[0])
        self.assertEqual(obs[1], exp[1])

        exp = ([], 0)
        obs = _get_rank([])
        assert_allclose(obs[0], exp[0])
        self.assertEqual(obs[1], exp[1])

    def test_get_rank_invalid_input(self):
        """Test the _get_rank function with invalid input."""
        vec = [1, "a", 3, 2.5, 3, 1]
        self.assertRaises(TypeError, _get_rank, vec)

        vec = [1, 2, {1: 2}, 2.5, 3, 1]
        self.assertRaises(TypeError, _get_rank, vec)

        vec = [1, 2, [23, 1], 2.5, 3, 1]
        self.assertRaises(TypeError, _get_rank, vec)

        vec = [1, 2, (1,), 2.5, 3, 1]
        self.assertRaises(TypeError, _get_rank, vec)

    def test_correlation(self):
        """Correlations and significance should match R's cor.test()"""
        x = [1, 2, 3, 5]
        y = [0, 0, 0, 0]
        z = [1, 1, 1, 1]
        a = [2, 4, 6, 8]
        b = [1.5, 1.4, 1.2, 1.1]
        c = [15, 10, 5, 20]

        bad = [1, 2, 3]  # originally gave r = 1.0000000002

        assert_allclose(pearson_correlation(x, x), (1, 0))
        assert_allclose(pearson_correlation(x, y), (0, 1), rtol=1e-3)
        assert_allclose(pearson_correlation(y, z), (0, 1))
        assert_allclose(
            pearson_correlation(x, a), (0.9827076, 0.01729), rtol=1e-4, atol=1e-6
        )
        assert_allclose(pearson_correlation(x, b), (-0.9621405, 0.03786), rtol=1e-4)
        assert_allclose(pearson_correlation(x, c), (0.3779645, 0.622), 1e-3)
        self.assertEqual(pearson_correlation(bad, bad), (1, 0))
        got = pearson_correlation(self.data1, self.data2, tails="low")
        assert_allclose(got, (-0.03760147385, 0.4589314864))

    def test_correlation_test_pearson(self):
        """Test correlation_test using pearson on valid input."""
        # These results were verified with R.

        # Test with non-default confidence level and permutations.
        obs = correlation_test(
            self.data1,
            self.data2,
            method="pearson",
            confidence_level=0.90,
            permutations=990,
        )
        assert_allclose(obs[:2], (-0.03760147, 0.91786297277172868), rtol=1e-6)
        self.assertEqual(len(obs[2]), 990)
        for r in obs[2]:
            self.assertTrue(r >= -1.0 and r <= 1.0)
        self.assertCorrectPValue(
            0.9,
            0.93,
            correlation_test,
            (self.data1, self.data2),
            {"method": "pearson", "confidence_level": 0.90, "permutations": 990},
            p_val_idx=3,
        )
        assert_allclose(obs[4], (-0.5779077, 0.5256224))

        # Test with non-default tail type.
        obs = correlation_test(
            self.data1,
            self.data2,
            method="pearson",
            confidence_level=0.90,
            permutations=990,
            tails="low",
        )
        assert_allclose(obs[:2], (-0.03760147, 0.45893148638586434), rtol=1e-6)
        self.assertEqual(len(obs[2]), 990)
        for r in obs[2]:
            self.assertTrue(r >= -1.0 and r <= 1.0)
        self.assertCorrectPValue(
            0.41,
            0.46,
            correlation_test,
            (self.data1, self.data2),
            {
                "method": "pearson",
                "confidence_level": 0.90,
                "permutations": 990,
                "tails": "low",
            },
            p_val_idx=3,
        )
        assert_allclose(obs[4], (-0.5779077, 0.5256224))

    def test_correlation_test_spearman(self):
        """Test correlation_test using spearman on valid input."""
        # This example taken from Wikipedia page:
        # http://en.wikipedia.org/wiki/Spearman's_rank_correlation_coefficient
        obs = correlation_test(self.data1, self.data2, method="spearman", tails="high")
        assert_allclose(obs[:2], (-0.17575757575757578, 0.686405827612))
        self.assertEqual(len(obs[2]), 999)
        for rho in obs[2]:
            self.assertTrue(rho >= -1.0 and rho <= 1.0)
        self.assertCorrectPValue(
            0.67,
            0.7,
            correlation_test,
            (self.data1, self.data2),
            {"method": "spearman", "tails": "high"},
            p_val_idx=3,
        )
        assert_allclose(obs[4], (-0.7251388558041697, 0.51034422964834503))

        # The p-value is off because the example uses a one-tailed test, while
        # we use a two-tailed test. Someone confirms the answer that we get
        # here for a two-tailed test:
        # http://stats.stackexchange.com/questions/22816/calculating-p-value-
        #     for-spearmans-rank-correlation-coefficient-example-on-wikip
        obs = correlation_test(self.data1, self.data2, method="spearman", tails=None)
        assert_allclose(obs[:2], (-0.17575757575757578, 0.62718834477648433))
        self.assertEqual(len(obs[2]), 999)
        for rho in obs[2]:
            self.assertTrue(rho >= -1.0 and rho <= 1.0)
        self.assertCorrectPValue(
            0.60,
            0.64,
            correlation_test,
            (self.data1, self.data2),
            {"method": "spearman", "tails": None},
            p_val_idx=3,
        )
        assert_allclose(obs[4], (-0.7251388558041697, 0.51034422964834503))

    def test_correlation_test_invalid_input(self):
        """Test correlation_test using invalid input."""
        self.assertRaises(
            ValueError, correlation_test, self.data1, self.data2, method="foo"
        )
        self.assertRaises(
            ValueError, correlation_test, self.data1, self.data2, tails="foo"
        )
        self.assertRaises(
            ValueError, correlation_test, self.data1, self.data2, permutations=-1
        )
        self.assertRaises(
            ValueError, correlation_test, self.data1, self.data2, confidence_level=-1
        )
        self.assertRaises(
            ValueError, correlation_test, self.data1, self.data2, confidence_level=1.1
        )
        self.assertRaises(
            ValueError, correlation_test, self.data1, self.data2, confidence_level=0
        )
        self.assertRaises(
            ValueError, correlation_test, self.data1, self.data2, confidence_level=0.0
        )
        self.assertRaises(
            ValueError, correlation_test, self.data1, self.data2, confidence_level=1
        )
        self.assertRaises(
            ValueError, correlation_test, self.data1, self.data2, confidence_level=1.0
        )

    def test_correlation_test_no_permutations(self):
        """Test correlation_test with no permutations."""
        # These results were verified with R.
        exp = (
            -0.2581988897471611,
            0.7418011102528389,
            [],
            None,
            (-0.97687328610475876, 0.93488023560400879),
        )
        obs = correlation_test([1, 2, 3, 4], [1, 2, 1, 1], permutations=0)
        for o, e in zip(obs, exp):
            if isinstance(e, type(None)):
                assert_equal(o, e)
            else:
                assert_allclose(o, e)

    def test_correlation_test_perfect_correlation(self):
        """Test correlation_test with perfectly-correlated input vectors."""
        # These results were verified with R.
        obs = correlation_test([1, 2, 3, 4], [1, 2, 3, 4])
        assert_allclose(obs[:2], (0.99999999999999978, 2.2204460492503131e-16))
        self.assertEqual(len(obs[2]), 999)
        for r in obs[2]:
            self.assertTrue(r >= -1.0 and r <= 1.0)
        self.assertCorrectPValue(
            0.06, 0.09, correlation_test, ([1, 2, 3, 4], [1, 2, 3, 4]), p_val_idx=3
        )
        assert_allclose(obs[4], (0.99999999999998879, 1.0))

    def test_correlation_test_small_obs(self):
        """Test correlation_test with a small number of observations."""
        # These results were verified with R.
        obs = correlation_test([1, 2, 3], [1, 2, 3])
        assert_allclose(obs[:2], (1.0, 0))
        self.assertEqual(len(obs[2]), 999)
        for r in obs[2]:
            self.assertTrue(r >= -1.0 and r <= 1.0)
        self.assertCorrectPValue(
            0.3, 0.4, correlation_test, ([1, 2, 3], [1, 2, 3]), p_val_idx=3
        )
        self.assertEqual(obs[4], (None, None))

        obs = correlation_test([1, 2, 3], [1, 2, 3], method="spearman")
        assert_allclose(obs[:2], (1.0, 0))
        self.assertEqual(len(obs[2]), 999)
        for r in obs[2]:
            self.assertTrue(r >= -1.0 and r <= 1.0)
        self.assertCorrectPValue(
            0.3,
            0.4,
            correlation_test,
            ([1, 2, 3], [1, 2, 3]),
            {"method": "spearman"},
            p_val_idx=3,
        )
        self.assertEqual(obs[4], (None, None))

    def test_correlation_matrix(self):
        """Correlations in matrix should match values from R"""
        a = [2, 4, 6, 8]
        b = [1.5, 1.4, 1.2, 1.1]
        c = [15, 10, 5, 20]
        m = correlation_matrix([a, b, c])
        assert_allclose(m[0, 0], [1.0])
        assert_allclose([m[1, 0], m[1, 1]], [correlation(b, a)[0], 1.0])
        assert_allclose(m[2], [correlation(c, a)[0], correlation(c, b)[0], 1.0])


class Ftest(TestCase):
    """Tests for the F test"""

    def test_f_value(self):
        """f_value: should calculate the correct F value if possible"""
        a = array([1, 3, 5, 7, 9, 8, 6, 4, 2])
        b = array([5, 4, 6, 3, 7, 6, 4, 5])
        self.assertEqual(f_value(a, b), (8, 7, 4.375))
        assert_allclose(f_value(b, a), (7, 8, 0.2285714), rtol=1e-6)
        too_short = array([4])
        self.assertRaises(ValueError, f_value, too_short, b)

    def test_f_two_sample(self):
        """f_two_sample should match values from R"""

        # The expected values in this test are obtained through R.
        # In R the F test is var.test(x,y) different alternative hypotheses
        # can be specified (two sided, less, or greater).
        # The vectors are random samples from a particular normal distribution
        # (mean and sd specified).

        # a: 50 elem, mean=0 sd=1
        a = [
            -0.70701689,
            -1.24788845,
            -1.65516470,
            0.10443876,
            -0.48526915,
            -0.71820656,
            -1.02603596,
            0.03975982,
            -2.23404324,
            -0.21509363,
            0.08438468,
            -0.01970062,
            -0.67907971,
            -0.89853667,
            1.11137131,
            0.05960496,
            -1.51172084,
            -0.79733957,
            -1.60040659,
            0.80530639,
            -0.81715836,
            -0.69233474,
            0.95750665,
            0.99576429,
            -1.61340216,
            -0.43572590,
            -1.50862327,
            0.92847551,
            -0.68382338,
            -1.12523522,
            -0.09147488,
            0.66756023,
            -0.87277588,
            -1.36539039,
            -0.11748707,
            -1.63632578,
            -0.31343078,
            -0.28176086,
            0.33854483,
            -0.51785630,
            2.25360559,
            -0.80761191,
            1.18983499,
            0.57080342,
            -1.44601700,
            -0.53906955,
            -0.01975266,
            -1.37147915,
            -0.31537616,
            0.26877544,
        ]

        # b: 50 elem, mean=0, sd=1.2
        b = [
            0.081418743,
            0.276571612,
            -1.864316504,
            0.675213612,
            -0.769202643,
            0.140372825,
            -1.426250184,
            0.058617884,
            -0.819287409,
            -0.007701916,
            -0.782722020,
            -0.285891593,
            0.661980419,
            0.383225191,
            0.622444946,
            -0.192446150,
            0.297150571,
            0.408896059,
            -0.167359383,
            -0.552381362,
            0.982168338,
            1.439730446,
            1.967616101,
            -0.579607307,
            1.095590943,
            0.240591302,
            -1.566937143,
            -0.199091349,
            -1.232983905,
            0.362378169,
            1.166061081,
            -0.604676222,
            -0.536560206,
            -0.303117595,
            1.519222792,
            -0.319146503,
            2.206220810,
            -0.566351124,
            -0.720397392,
            -0.452001377,
            0.250890097,
            0.320685395,
            -1.014632725,
            -3.010346273,
            -1.703955054,
            0.592587381,
            -1.237451255,
            0.172243366,
            -0.452641122,
            -0.982148581,
        ]

        # c: 60 elem, mean=5, sd=1
        c = [
            4.654329,
            5.242129,
            6.272640,
            5.781779,
            4.391241,
            3.800752,
            4.559463,
            4.318922,
            3.243020,
            5.121280,
            4.126385,
            5.541131,
            4.777480,
            5.646913,
            6.972584,
            3.817172,
            6.128700,
            4.731467,
            6.762068,
            5.082983,
            5.298511,
            5.491125,
            4.532369,
            4.265552,
            5.697317,
            5.509730,
            2.935704,
            4.507456,
            3.786794,
            5.548383,
            3.674487,
            5.536556,
            5.297847,
            2.439642,
            4.759836,
            5.114649,
            5.986774,
            4.517485,
            4.579208,
            4.579374,
            2.502890,
            5.190955,
            5.983194,
            6.766645,
            4.905079,
            4.214273,
            3.950364,
            6.262393,
            8.122084,
            6.330007,
            4.767943,
            5.194029,
            3.503136,
            6.039079,
            4.485647,
            6.116235,
            6.302268,
            3.596693,
            5.743316,
            6.860152,
        ]

        # d: 30 elem, mean=0, sd =0.05
        d = [
            0.104517366,
            0.023039678,
            0.005579091,
            0.052928250,
            0.020724823,
            -0.060823243,
            -0.019000890,
            -0.064133996,
            -0.016321594,
            -0.008898334,
            -0.027626992,
            -0.051946186,
            0.085269587,
            -0.031190678,
            0.065172938,
            -0.054628573,
            0.019257306,
            -0.032427056,
            -0.058767356,
            0.030927400,
            0.052247357,
            -0.042954937,
            0.031842104,
            0.094130522,
            -0.024828465,
            0.011320453,
            -0.016195062,
            0.015631245,
            -0.050335598,
            -0.031658335,
        ]

        a, b, c, d = list(map(array, [a, b, c, d]))
        self.assertEqual(list(map(len, [a, b, c, d])), [50, 50, 60, 30])

        # allowed error. This big, because results from R
        # are rounded at 4 decimals
        error = 1e-3

        assert_allclose(f_two_sample(a, a), (49, 49, 1, 1), rtol=error)
        assert_allclose(f_two_sample(a, b), (49, 49, 0.8575, 0.5925), rtol=error)
        assert_allclose(f_two_sample(b, a), (49, 49, 1.1662, 0.5925), rtol=error)
        assert_allclose(
            f_two_sample(a, b, tails="low"), (49, 49, 0.8575, 0.2963), rtol=error
        )
        assert_allclose(
            f_two_sample(a, b, tails="high"), (49, 49, 0.8575, 0.7037), rtol=error
        )
        assert_allclose(f_two_sample(a, c), (49, 59, 0.6587, 0.1345), rtol=error)
        # p value very small, so first check df's and F value
        assert_allclose(
            f_two_sample(d, a, tails="low")[0:3],
            (29, 49, 0.0028),
            rtol=error,
            atol=1e-4,
        )
        assert f_two_sample(d, a, tails="low")[3] < 2.2e-16  # p value

    def test_MonteCarloP(self):
        """MonteCarloP calcs a p-value from a val and list of random vals"""
        val = 3.0
        random_vals = [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]

        # test for "high" tail (larger values than expected by chance)
        p_val = MonteCarloP(val, random_vals, "high")
        self.assertEqual(p_val, 0.7)

        # test for "low" tail (smaller values than expected by chance)
        p_val = MonteCarloP(val, random_vals, "low")
        self.assertEqual(p_val, 0.4)


class MannWhitneyTests(TestCase):
    """check accuracy of Mann-Whitney implementation"""

    x = list(
        map(
            int,
            "104 109 112 114 116 118 118 119 121 123 125 126"
            " 126 128 128 128".split(),
        )
    )
    y = list(map(int, "100 105 107 107 108 111 116 120 121 123".split()))

    def test_mw_test(self):
        """mann-whitney test results should match Sokal & Rohlf"""
        U, p = mw_test(self.x, self.y)
        assert_allclose(U, 123.5)
        self.assertTrue(0.02 <= p <= 0.05)

    def test_mw_boot(self):
        """excercising the Monte-carlo variant of mann-whitney"""
        U, p = mw_boot(self.x, self.y, 10)
        assert_allclose(U, 123.5)
        self.assertTrue(0 <= p <= 0.5)


class KendallTests(TestCase):
    """check accuracy of Kendall tests against values from R"""

    def do_test(self, x, y, alt_expecteds):
        """conducts the tests for each alternate hypothesis against expecteds"""
        for alt, exp_p, exp_tau in alt_expecteds:
            tau, p_val = kendall_correlation(x, y, alt=alt, warn=False)
            assert_allclose(tau, exp_tau, rtol=1e-3)
            assert_allclose(p_val, exp_p, rtol=1e-3)

    def test_exact_calcs(self):
        """calculations of exact probabilities should match R"""
        x = (44.4, 45.9, 41.9, 53.3, 44.7, 44.1, 50.7, 45.2, 60.1)
        y = (2.6, 3.1, 2.5, 5.0, 3.6, 4.0, 5.2, 2.8, 3.8)
        expecteds = [
            ["gt", 0.05972, 0.4444444],
            ["lt", 0.9624, 0.4444444],
            ["ts", 0.1194, 0.4444444],
        ]
        self.do_test(x, y, expecteds)

    def test_with_ties(self):
        """tied values calculated from normal approx"""
        # R example with ties in x
        x = (44.4, 45.9, 41.9, 53.3, 44.4, 44.1, 50.7, 45.2, 60.1)
        y = (2.6, 3.1, 2.5, 5.0, 3.6, 4.0, 5.2, 2.8, 3.8)
        expecteds = [  # ["gt", 0.05793, 0.4225771],
            ["lt", 0.942, 0.4225771],
            ["ts", 0.1159, 0.4225771],
        ]
        self.do_test(x, y, expecteds)

        # R example with ties in y
        x = (44.4, 45.9, 41.9, 53.3, 44.7, 44.1, 50.7, 45.2, 60.1)
        y = (2.6, 3.1, 2.5, 5.0, 3.1, 4.0, 5.2, 2.8, 3.8)
        expecteds = [
            ["gt", 0.03737, 0.4789207],
            ["lt", 0.9626, 0.4789207],
            ["ts", 0.07474, 0.4789207],
        ]
        self.do_test(x, y, expecteds)
        # R example with ties in x and y
        x = (44.4, 45.9, 41.9, 53.3, 44.7, 44.1, 50.7, 44.4, 60.1)
        y = (2.6, 3.6, 2.5, 5.0, 3.6, 4.0, 5.2, 2.8, 3.8)
        expecteds = [
            ["gt", 0.02891, 0.5142857],
            ["lt", 0.971, 0.5142857],
            ["ts", 0.05782, 0.5142857],
        ]
        self.do_test(x, y, expecteds)

    def test_bigger_vectors(self):
        """docstring for test_bigger_vectors"""
        # q < expansion
        x = (
            0.118583104633,
            0.227860069338,
            0.143856130991,
            0.935362617582,
            0.0471303856799,
            0.659819202174,
            0.739247965907,
            0.268929000278,
            0.848250568194,
            0.307764819102,
            0.733949480141,
            0.271662210481,
            0.155903098872,
        )
        y = (
            0.749762144455,
            0.407571703468,
            0.934176427266,
            0.188638794706,
            0.184844781493,
            0.391485553856,
            0.735504815302,
            0.363655952442,
            0.18489971978,
            0.851075466765,
            0.139932273818,
            0.333675110224,
            0.570250937033,
        )
        expecteds = [
            ["gt", 0.9183, -0.2820513],
            ["lt", 0.1022, -0.2820513],
            ["ts", 0.2044, -0.2820513],
        ]
        self.do_test(x, y, expecteds)
        # q > expansion
        x = (
            0.2602556958,
            0.441506392849,
            0.930624643531,
            0.728461775775,
            0.234341774892,
            0.725677256368,
            0.354788882728,
            0.475882541956,
            0.347533553428,
            0.608578046857,
            0.144697962102,
            0.784502692164,
            0.872607603407,
        )
        y = (
            0.753056395718,
            0.454332072011,
            0.791882395707,
            0.622853579015,
            0.127030232518,
            0.232086215578,
            0.586604349918,
            0.0139051260749,
            0.579079370051,
            0.0550643809812,
            0.94798878249,
            0.318410679439,
            0.86725134615,
        )
        expecteds = [
            ["gt", 0.4762, 0.02564103],
            ["lt", 0.5711, 0.02564103],
            ["ts", 0.9524, 0.02564103],
        ]
        self.do_test(x, y, expecteds)


class TestDistMatrixPermutationTest(TestCase):
    """Tests of distance_matrix_permutation_test"""

    def setUp(self):
        """sets up variables for testing"""
        self.matrix = array(
            [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]]
        )
        self.cells = [(0, 1), (1, 3)]
        self.cells2 = [(0, 2), (2, 3)]

    def test_get_ltm_cells(self):
        "get_ltm_cells converts indices to be below the diagonal"
        cells = [(0, 0), (0, 1), (0, 2), (1, 0), (1, 1), (1, 2), (2, 0), (2, 1), (2, 2)]
        result = get_ltm_cells(cells)
        self.assertEqual(result, [(1, 0), (2, 0), (2, 1)])
        cells = [(0, 1), (0, 2)]
        result = get_ltm_cells(cells)
        self.assertEqual(result, [(1, 0), (2, 0)])

    def test_get_values_from_matrix(self):
        """get_values_from_matrix returns the special and other values from matrix"""
        matrix = self.matrix
        cells = [(1, 0), (0, 1), (2, 0), (2, 1)]
        # test that works for a symmetric matrix
        cells_sym = get_ltm_cells(cells)
        special_vals, other_vals = get_values_from_matrix(
            matrix, cells_sym, cells2=None, is_symmetric=True
        )
        special_vals.sort()
        other_vals.sort()
        self.assertEqual(special_vals, [5, 9, 10])
        self.assertEqual(other_vals, [13, 14, 15])

        # test that work for a non symmetric matrix
        special_vals, other_vals = get_values_from_matrix(
            matrix, cells, cells2=None, is_symmetric=False
        )
        special_vals.sort()
        other_vals.sort()
        self.assertEqual(special_vals, [2, 5, 9, 10])
        self.assertEqual(other_vals, [1, 3, 4, 6, 7, 8, 11, 12, 13, 14, 15, 16])

        # test that works on a symmetric matrix when cells2 is defined
        cells2 = [(3, 0), (3, 2), (0, 3)]
        cells2_sym = get_ltm_cells(cells2)
        special_vals, other_vals = get_values_from_matrix(
            matrix, cells_sym, cells2=cells2_sym, is_symmetric=True
        )
        special_vals.sort()
        other_vals.sort()
        self.assertEqual(special_vals, [5, 9, 10])
        self.assertEqual(other_vals, [13, 15])

        # test that works when cells2 is defined and not symmetric
        special_vals, other_vals = get_values_from_matrix(
            matrix, cells, cells2=cells2, is_symmetric=False
        )
        special_vals.sort()
        other_vals.sort()
        self.assertEqual(special_vals, [2, 5, 9, 10])
        self.assertEqual(other_vals, [4, 13, 15])

    def test_distance_matrix_permutation_test_non_symmetric(self):
        """evaluate empirical p-values for a non symmetric matrix

        To test the empirical p-values, we look at a simple 3x3 matrix
         b/c it is easy to see what t score every permutation will
         generate -- there's only 6 permutations.
         Running dist_matrix_test with n=1000, we expect that each
         permutation will show up 160 times, so we know how many
         times to expect to see more extreme t scores. We therefore
         know what the empirical p-values will be. (n=1000 was chosen
         empirically -- smaller values seem to lead to much more frequent
         random failures.)


        """

        def make_result_list(*args, **kwargs):
            return [
                distance_matrix_permutation_test(*args, **kwargs)[2] for i in range(10)
            ]

        m = arange(9).reshape((3, 3))
        n = 100
        # looks at each possible permutation n times --
        # compare first row to rest
        r = make_result_list(m, [(0, 0), (0, 1), (0, 2)], n=n, is_symmetric=False)
        self.assertTrue(similar_means(r, 0.0 / 6.0))
        r = make_result_list(
            m, [(0, 0), (0, 1), (0, 2)], n=n, is_symmetric=False, tails="high"
        )
        self.assertTrue(similar_means(r, 4.0 / 6.0))
        r = make_result_list(
            m, [(0, 0), (0, 1), (0, 2)], n=n, is_symmetric=False, tails="low"
        )
        self.assertTrue(similar_means(r, 0.0 / 6.0))

        # looks at each possible permutation n times --
        # compare last row to rest
        r = make_result_list(m, [(2, 0), (2, 1), (2, 2)], n=n, is_symmetric=False)
        self.assertTrue(similar_means(r, 0.0 / 6.0))
        r = make_result_list(
            m, [(2, 0), (2, 1), (2, 2)], n=n, is_symmetric=False, tails="high"
        )
        self.assertTrue(similar_means(r, 0.0 / 6.0))
        r = make_result_list(
            m, [(2, 0), (2, 1), (2, 2)], n=n, is_symmetric=False, tails="low"
        )
        self.assertTrue(similar_means(r, 4.0 / 6.0))

    def test_distance_matrix_permutation_test_symmetric(self):
        """evaluate empirical p-values for symmetric matrix

        See test_distance_matrix_permutation_test_non_symmetric
        doc string for a description of how this test works.

        """

        def make_result_list(*args, **kwargs):
            return [distance_matrix_permutation_test(*args)[2] for i in range(10)]

        m = array([[0, 1, 3], [1, 2, 4], [3, 4, 5]])
        # looks at each possible permutation n times --
        # compare first row to rest
        n = 100

        # looks at each possible permutation n times --
        # compare first row to rest
        r = make_result_list(m, [(0, 0), (0, 1), (0, 2)], n=n)
        self.assertTrue(similar_means(r, 2.0 / 6.0))
        r = make_result_list(m, [(0, 0), (0, 1), (0, 2)], n=n, tails="high")
        self.assertTrue(similar_means(r, 0.77281447417149496, 0))
        r = make_result_list(m, [(0, 0), (0, 1), (0, 2)], n=n, tails="low")
        self.assertTrue(similar_means(r, 2.0 / 6.0))

        # The following lines are not part of the test code, but are useful in
        # figuring out what t-scores all of the permutations will yield.
        # permutes = [[0, 1, 2], [0, 2, 1], [1, 0, 2],\
        # [1, 2, 0], [2, 0, 1], [2, 1, 0]]
        # results = []
        # for p in permutes:
        #    p_m = permute_2d(m,p)
        #    results.append(t_two_sample(\
        #     [p_m[0,1],p_m[0,2]],[p_m[2,1]],tails='high'))
        # print results

    def test_distance_matrix_permutation_test_alt_stat(self):
        def fake_stat_test(a, b, tails=None):
            return 42.0, 42.0

        m = array([[0, 1, 3], [1, 2, 4], [3, 4, 5]])
        self.assertEqual(
            distance_matrix_permutation_test(
                m, [(0, 0), (0, 1), (0, 2)], n=5, f=fake_stat_test
            ),
            (42.0, 42.0, 0.0),
        )

    def test_distance_matrix_permutation_test_return_scores(self):
        """return_scores=True functions as expected"""
        # use alt statistical test to make results simple
        def fake_stat_test(a, b, tails=None):
            return 42.0, 42.0

        m = array([[0, 1, 3], [1, 2, 4], [3, 4, 5]])
        self.assertEqual(
            distance_matrix_permutation_test(
                m, [(0, 0), (0, 1), (0, 2)], n=5, f=fake_stat_test, return_scores=True
            ),
            (42.0, 42.0, 0.0, [42.0] * 5),
        )

    def test_ANOVA_one_way(self):
        """ANOVA one way returns same values as ANOVA on a stats package"""
        g1 = NumberCounter([10.0, 11.0, 10.0, 5.0, 6.0])
        g2 = NumberCounter([1.0, 2.0, 3.0, 4.0, 1.0, 2.0])
        g3 = NumberCounter([6.0, 7.0, 5.0, 6.0, 7.0])
        i = [g1, g2, g3]
        dfn, dfd, F, between_MS, within_MS, group_means, prob = ANOVA_one_way(i)
        self.assertEqual(dfn, 2)
        self.assertEqual(dfd, 13)
        assert_allclose(F, 18.565450643776831)
        assert_allclose(between_MS, 55.458333333333343)
        assert_allclose(within_MS, 2.9871794871794868)
        assert_allclose(
            group_means, [8.4000000000000004, 2.1666666666666665, 6.2000000000000002]
        )
        assert_allclose(prob, 0.00015486238993089464)


# execute tests if called from command line
if __name__ == "__main__":
    main()



from collections import defaultdict, namedtuple
from numbers import Number

import numpy

from numpy import array, diag, dot, eye, float64, int32, log, sqrt, zeros
from numpy.linalg import det, inv

from cogent3.core.moltype import DNA, RNA, get_moltype
from cogent3.util.dict_array import DictArray
from cogent3.util.misc import get_object_provenance
from cogent3.util.progress_display import display_wrap

from .pairwise_distance_numba import fill_diversity_matrix


__author__ = "Gavin Huttley, Yicheng Zhu and Ben Kaehler"
__copyright__ = "Copyright 2007-2022, The Cogent Project"
__credits__ = ["Gavin Huttley", "Yicheng Zhu", "Ben Kaehler"]
__license__ = "BSD-3"
__version__ = "2023.2.12a1"
__maintainer__ = "Gavin Huttley"
__email__ = "gavin.huttley@anu.edu.au"
__status__ = "Alpha"  # pending addition of protein distance metrics


def _same_moltype(ref, query):
    """if ref and query have the same states"""
    return set(ref) == set(query)


def get_pyrimidine_indices(moltype):
    """returns pyrimidine indices for the moltype"""
    states = list(moltype)
    if _same_moltype(RNA, moltype):
        return list(map(states.index, "CU"))
    elif _same_moltype(DNA, moltype):
        return list(map(states.index, "CT"))
    else:
        raise RuntimeError("Non-nucleic acid MolType")


def get_purine_indices(moltype):
    """returns purine indices for the moltype"""
    states = list(moltype)
    if not _same_moltype(RNA, moltype) and not _same_moltype(DNA, moltype):
        raise RuntimeError("Non-nucleic acid MolType")

    return list(map(states.index, "AG"))


def get_matrix_diff_coords(indices):
    """returns coordinates for off diagonal elements"""
    return [(i, j) for i in indices for j in indices if i != j]


def get_moltype_index_array(moltype, invalid=-9):
    """returns the index array for a molecular type"""
    canonical_chars = list(moltype)
    # maximum ordinal for an allowed character, this defines the length of
    # the required numpy array
    max_ord = max(list(map(ord, list(moltype.All.keys()))))
    char_to_index = zeros(max_ord + 1, int32)
    # all non canonical_chars are ``invalid''
    char_to_index.fill(invalid)

    for i in range(len(canonical_chars)):
        c = canonical_chars[i]
        o = ord(c)
        char_to_index[o] = i

    return char_to_index


def seq_to_indices(seq, char_to_index):
    """returns an array with sequence characters replaced by their index"""
    ords = list(map(ord, seq))
    return char_to_index.take(ords)


def _fill_diversity_matrix(matrix, seq1, seq2):
    """fills the diversity matrix for valid positions.

    Assumes the provided sequences have been converted to indices with
    invalid characters being negative numbers (use get_moltype_index_array
    plus seq_to_indices)."""
    paired = array([seq1, seq2]).T
    paired = paired[paired.min(axis=1) >= 0]
    for i in range(len(paired)):
        matrix[paired[i][0], paired[i][1]] += 1


def _hamming(matrix):
    """computes the edit distance
    Parameters
    ----------
    matrix : array
        2D numpy array of counts
    Returns
    -------
    total of the matrix, the proportion of changes, hamming distance, variance
    (the variance calculation is not yet implemented)
    """
    # todo implement the estimate of the variance
    invalid = None, None, None, None
    total = matrix.sum()
    dist = total - diag(matrix).sum()
    if total == 0:
        return invalid

    p = dist / total

    return total, p, dist, None


def _jc69_from_matrix(matrix):
    """computes JC69 stats from a diversity matrix"""
    invalid = None, None, None, None
    total = matrix.sum()
    diffs = total - diag(matrix).sum()
    if total == 0:
        return invalid

    p = diffs / total
    if p >= 0.75:  # cannot take log
        return invalid

    factor = 1 - (4 / 3) * p

    dist = -3.0 * log(factor) / 4
    var = p * (1 - p) / (factor * factor * total)
    return total, p, dist, var


def _tn93_from_matrix(
    matrix, freqs, pur_indices, pyr_indices, pur_coords, pyr_coords, tv_coords
):
    invalid = None, None, None, None

    total = matrix.sum()
    freqs = matrix.sum(axis=0) + matrix.sum(axis=1)
    freqs /= 2 * total

    if total == 0:
        return invalid

    p = matrix.take(pur_coords + pyr_coords + tv_coords).sum() / total

    freq_purs = freqs.take(pur_indices).sum()
    prod_purs = freqs.take(pur_indices).prod()
    freq_pyrs = freqs.take(pyr_indices).sum()
    prod_pyrs = freqs.take(pyr_indices).prod()

    # purine transition diffs
    pur_ts_diffs = matrix.take(pur_coords).sum()
    pur_ts_diffs /= total
    # pyr transition  diffs
    pyr_ts_diffs = matrix.take(pyr_coords).sum()
    pyr_ts_diffs /= total
    # transversions
    tv_diffs = matrix.take(tv_coords).sum() / total

    coeff1 = 2 * prod_purs / freq_purs
    coeff2 = 2 * prod_pyrs / freq_pyrs
    coeff3 = 2 * (
        freq_purs * freq_pyrs
        - (prod_purs * freq_pyrs / freq_purs)
        - (prod_pyrs * freq_purs / freq_pyrs)
    )

    term1 = 1 - pur_ts_diffs / coeff1 - tv_diffs / (2 * freq_purs)
    term2 = 1 - pyr_ts_diffs / coeff2 - tv_diffs / (2 * freq_pyrs)
    term3 = 1 - tv_diffs / (2 * freq_purs * freq_pyrs)

    if term1 <= 0 or term2 <= 0 or term3 <= 0:  # log will fail
        return invalid

    dist = -coeff1 * log(term1) - coeff2 * log(term2) - coeff3 * log(term3)
    v1 = 1 / term1
    v2 = 1 / term2
    v3 = 1 / term3
    v4 = (
        (coeff1 * v1 / (2 * freq_purs))
        + (coeff2 * v2 / (2 * freq_pyrs))
        + (coeff3 * v3 / (2 * freq_purs * freq_pyrs))
    )
    var = (
        v1 ** 2 * pur_ts_diffs
        + v2 ** 2 * pyr_ts_diffs
        + v4 ** 2 * tv_diffs
        - (v1 * pur_ts_diffs + v2 * pyr_ts_diffs + v4 * tv_diffs) ** 2
    )
    var /= total

    return total, p, dist, var


def _logdetcommon(matrix):
    invalid = (None,) * 5

    total = matrix.sum()
    diffs = total - matrix.diagonal().sum()
    if total == 0:
        return invalid

    p = diffs / total

    if diffs == 0:  # seqs indentical
        return invalid

    # we replace the missing diagonal states with a frequency of 0.5,
    # then normalise
    frequency = matrix.copy()
    frequency[(frequency == 0) * eye(*matrix.shape, dtype=bool)] = 0.5
    frequency /= frequency.sum()

    if det(frequency) <= 0:  # if the result is nan
        return invalid

    # the inverse matrix of frequency, every element is squared
    M_matrix = inv(frequency) ** 2
    freqs = [frequency.sum(axis=axis) for axis in (0, 1)]
    var_term = dot(M_matrix, frequency).diagonal().sum()

    return total, p, frequency, freqs, var_term


def _paralinear(matrix):
    """the paralinear distance from a diversity matrix"""

    invalid = (None,) * 4

    total, p, frequency, freqs, var_term = _logdetcommon(matrix)
    if frequency is None:
        return invalid

    r = matrix.shape[0]
    d_xy = -log(det(frequency) / sqrt((freqs[0] * freqs[1]).prod())) / r
    var = (var_term - (1 / sqrt(freqs[0] * freqs[1])).sum()) / (r ** 2 * total)

    return total, p, d_xy, var


def _logdet(matrix, use_tk_adjustment=True):
    """returns the LogDet from a diversity matrix

    Parameters
    ----------
    use_tk_adjustment
        when True, unequal state frequencies are allowed

    """

    invalid = (None,) * 4

    total, p, frequency, freqs, var_term = _logdetcommon(matrix)
    if frequency is None:
        return invalid

    r = matrix.shape[0]
    if use_tk_adjustment:
        coeff = (sum(sum(freqs) ** 2) / 4 - 1) / (r - 1)
        d_xy = coeff * log(det(frequency) / sqrt((freqs[0] * freqs[1]).prod()))
        var = None
    else:
        d_xy = -log(det(frequency)) / r - log(r)
        var = (var_term / r ** 2 - 1) / total

    return total, p, d_xy, var


def _number_formatter(template):
    """flexible number formatter"""

    def call(val):
        try:
            result = template % val
        except TypeError:
            result = val
        return result

    return call


Stats = namedtuple("Stats", ["length", "fraction_variable", "dist", "variance"])


def _make_stat_table(stats, names, **kwargs):
    from cogent3.util.table import Table

    header = [r"Seq1 \ Seq2"] + names
    rows = zeros((len(names), len(names)), dtype="O")
    for i in range(len(names) - 1):
        n1 = names[i]
        for j in range(i + 1, len(names)):
            n2 = names[j]
            val = stats[(n1, n2)]
            rows[i, j] = val
            rows[j, i] = val
    rows = rows.tolist()
    for i in range(len(names)):
        rows[i].insert(0, names[i])

    return Table(
        header=header, data=rows, index_name=r"Seq1 \ Seq2", missing_data="*", **kwargs
    )


class _PairwiseDistance(object):
    """base class for computing pairwise distances"""

    valid_moltypes = ()

    def __init__(self, moltype, invalid=-9, alignment=None, invalid_raises=False):
        super(_PairwiseDistance, self).__init__()
        moltype = get_moltype(moltype)
        if moltype.label not in self.valid_moltypes:
            name = self.__class__.__name__
            msg = (
                f"Invalid moltype for {name}: '{moltype.label}' not "
                f"in {self.valid_moltypes}"
            )
            raise ValueError(msg)

        self.moltype = moltype
        self.char_to_indices = get_moltype_index_array(moltype, invalid=invalid)
        self._dim = len(list(moltype))
        self._dists = None
        self._dupes = None
        self._duped = None
        self._invalid_raises = invalid_raises

        self.names = None
        self.indexed_seqs = None

        if alignment is not None:
            self._convert_seqs_to_indices(alignment)

        self._func_args = []

    def _convert_seqs_to_indices(self, alignment):
        assert isinstance(
            alignment.moltype, type(self.moltype)
        ), "Alignment does not have correct MolType"

        self._dists = {}
        self.names = alignment.names[:]
        indexed_seqs = []
        for name in self.names:
            seq = alignment.get_gapped_seq(name)
            indexed = seq_to_indices(str(seq), self.char_to_indices)
            indexed_seqs.append(indexed)

        self.indexed_seqs = array(indexed_seqs)

    @property
    def duplicated(self):
        """returns mapping of IDs to duplicates as {id:[dupe1, ..], },
        or None"""
        return self._duped

    @staticmethod
    def func():
        pass  # over ride in subclasses

    @display_wrap
    def run(self, alignment=None, ui=None):
        """computes the pairwise distances"""
        self._dupes = None
        self._duped = None

        dupes = set()
        duped = defaultdict(list)

        if alignment is not None:
            self._convert_seqs_to_indices(alignment)

        names = self.names[:]
        matrix = zeros((self._dim, self._dim), float64)
        off_diag = [
            (i, j) for i in range(self._dim) for j in range(self._dim) if i != j
        ]
        off_diag = tuple([tuple(a) for a in zip(*off_diag)])

        done = 0.0
        to_do = (len(names) * len(names) - 1) / 2
        for i in range(len(names) - 1):
            if i in dupes:
                continue

            name_1 = names[i]
            s1 = self.indexed_seqs[i]
            for j in range(i + 1, len(names)):
                if j in dupes:
                    continue

                name_2 = names[j]
                ui.display(f"{name_1} vs {name_2}", done / to_do)
                done += 1
                matrix.fill(0)
                s2 = self.indexed_seqs[j]
                fill_diversity_matrix(matrix, s1, s2)
                if not (matrix[off_diag] > 0).any():
                    # j is a duplicate of i
                    dupes.update([j])
                    duped[i].append(j)
                    continue

                total, p, dist, var = self.func(matrix, *self._func_args)
                if self._invalid_raises and not isinstance(dist, Number):
                    msg = f"distance could not be calculated for {name_1} - {name_2}"
                    raise ArithmeticError(msg)
                result = Stats(total, p, dist, var)
                self._dists[(name_1, name_2)] = result
                self._dists[(name_2, name_1)] = result

        self._dupes = [names[i] for i in dupes] or None
        if duped:
            self._duped = {}
            for k, v in duped.items():
                key = names[k]
                vals = [names[i] for i in v]
                self._duped[key] = vals

            # clean the distances so only unique seqs included
            remove = set(self._dupes)
            keys = list(self._dists.keys())
            for key in keys:
                if set(key) & remove:
                    del self._dists[key]

    __call__ = run

    def get_pairwise_distances(self, include_duplicates=True):
        """returns a matrix of pairwise distances.

        Parameters
        ----------
        include_duplicates : bool
            all seqs included in the distances, otherwise only unique sequences
            are included.
        """
        if self._dists is None:
            return None

        dists = {k: self._dists[k].dist for k in self._dists}
        if include_duplicates:
            dists = self._expand(dists)

        return DistanceMatrix(dists)

    def _expand(self, pwise):
        """returns a pwise statistic dict that includes duplicates"""
        if not self.duplicated:
            # no duplicates, nothing to do
            return pwise

        redundants = {}
        for k in self.duplicated:
            for r in self.duplicated[k]:
                redundants[r] = k

        names = self.names[:]
        for add, alias in redundants.items():
            for name in names:
                if name == add:
                    continue
                if name == alias:
                    val = 0
                else:
                    val = pwise.get((alias, name), None)
                pwise[(add, name)] = pwise[(name, add)] = val

        return pwise

    @property
    def dists(self):
        if self._dists is None:
            return None

        return self.get_pairwise_distances(include_duplicates=True)

    @property
    def stderr(self):
        if self._dists is None:
            return None

        stats = {k: sqrt(self._dists[k].variance) for k in self._dists}
        stats = self._expand(stats)
        kwargs = dict(title="Standard Error of Pairwise Distances", digits=4)
        return _make_stat_table(stats, self.names, **kwargs)

    @property
    def variances(self):
        if self._dists is None:
            return None

        stats = {k: self._dists[k].variance for k in self._dists}
        stats = self._expand(stats)
        kwargs = dict(title="Variances of Pairwise Distances", digits=4)
        t = _make_stat_table(stats, self.names, **kwargs)
        var_formatter = _number_formatter("%.2e")
        for name in self.names:
            t.format_column(name, var_formatter)
        return t

    @property
    def proportions(self):
        if self._dists is None:
            return None

        stats = {k: self._dists[k].fraction_variable for k in self._dists}
        stats = self._expand(stats)
        kwargs = dict(title="Proportion variable sites", digits=4)
        return _make_stat_table(stats, self.names, **kwargs)

    @property
    def lengths(self):
        if self._dists is None:
            return None

        stats = {k: self._dists[k].length for k in self._dists}
        stats = self._expand(stats)
        kwargs = dict(title="Pairwise Aligned Lengths", digits=0)
        return _make_stat_table(stats, self.names, **kwargs)


class HammingPair(_PairwiseDistance):
    """Hamming distance calculator for pairwise alignments"""

    valid_moltypes = ("dna", "rna", "protein", "text", "bytes")

    def __init__(self, moltype="text", *args, **kwargs):
        """states: the valid sequence states"""
        super(HammingPair, self).__init__(moltype, *args, **kwargs)
        self.func = _hamming


class PercentIdentityPair(_PairwiseDistance):
    """Percent identity distance calculator for pairwise alignments"""

    valid_moltypes = ("dna", "rna", "protein", "text", "bytes")

    def __init__(self, moltype="text", *args, **kwargs):
        """states: the valid sequence states"""
        super(PercentIdentityPair, self).__init__(moltype, *args, **kwargs)
        self.func = _hamming

    def get_pairwise_distances(self, include_duplicates=True):
        """returns a matrix of pairwise distances.

        Parameters
        ----------
        include_duplicates : bool
            all seqs included in the distances, otherwise only unique sequences
            are included.
        """
        if self._dists is None:
            return None

        dists = {k: self._dists[k].fraction_variable for k in self._dists}
        if include_duplicates:
            dists = self._expand(dists)

        return DistanceMatrix(dists)


class _NucleicSeqPair(_PairwiseDistance):
    """base class pairwise distance calculator for nucleic acid seqs"""

    valid_moltypes = ("dna", "rna")

    def __init__(self, moltype="dna", *args, **kwargs):
        super(_NucleicSeqPair, self).__init__(moltype, *args, **kwargs)
        if not _same_moltype(DNA, self.moltype) and not _same_moltype(
            RNA, self.moltype
        ):
            raise RuntimeError("Invalid MolType for this metric")


class JC69Pair(_NucleicSeqPair):
    """JC69 distance calculator for pairwise alignments"""

    def __init__(self, moltype="dna", *args, **kwargs):
        """states: the valid sequence states"""
        super(JC69Pair, self).__init__(moltype, *args, **kwargs)
        self.func = _jc69_from_matrix


class TN93Pair(_NucleicSeqPair):
    """TN93 calculator for pairwise alignments"""

    def __init__(self, moltype="dna", *args, **kwargs):
        """states: the valid sequence states"""
        super(TN93Pair, self).__init__(moltype, *args, **kwargs)
        self._freqs = zeros(self._dim, float64)
        self.pur_indices = get_purine_indices(self.moltype)
        self.pyr_indices = get_pyrimidine_indices(self.moltype)

        # matrix coordinates
        self.pyr_coords = get_matrix_diff_coords(self.pyr_indices)
        self.pur_coords = get_matrix_diff_coords(self.pur_indices)

        self.tv_coords = get_matrix_diff_coords(list(range(self._dim)))
        for coord in self.pur_coords + self.pyr_coords:
            self.tv_coords.remove(coord)

        # flattened
        self.pyr_coords = [i * 4 + j for i, j in self.pyr_coords]
        self.pur_coords = [i * 4 + j for i, j in self.pur_coords]
        self.tv_coords = [i * 4 + j for i, j in self.tv_coords]

        self.func = _tn93_from_matrix
        self._func_args = [
            self._freqs,
            self.pur_indices,
            self.pyr_indices,
            self.pur_coords,
            self.pyr_coords,
            self.tv_coords,
        ]


class LogDetPair(_PairwiseDistance):
    """computes logdet distance between sequence pairs"""

    valid_moltypes = ("dna", "rna", "protein")

    def __init__(self, moltype="dna", use_tk_adjustment=True, *args, **kwargs):
        """Arguments:
        - moltype: string or moltype instance (must be dna or rna)
        - use_tk_adjustment: use the correction of Tamura and Kumar 2002
        """
        super(LogDetPair, self).__init__(moltype, *args, **kwargs)
        self.func = _logdet
        self._func_args = [use_tk_adjustment]

    def run(self, use_tk_adjustment=None, *args, **kwargs):
        if use_tk_adjustment is not None:
            self._func_args = [use_tk_adjustment]

        super(LogDetPair, self).run(*args, **kwargs)


class ParalinearPair(_PairwiseDistance):
    """computes the paralinear distance (Lake 1994) between sequence pairs"""

    valid_moltypes = ("dna", "rna", "protein")

    def __init__(self, moltype="dna", *args, **kwargs):
        super(ParalinearPair, self).__init__(moltype, *args, **kwargs)
        self.func = _paralinear


_calculators = {
    "paralinear": ParalinearPair,
    "logdet": LogDetPair,
    "jc69": JC69Pair,
    "tn93": TN93Pair,
    "hamming": HammingPair,
    "percent": PercentIdentityPair,
}


def get_distance_calculator(name, *args, **kwargs):
    """returns a pairwise distance calculator

    name is converted to lower case"""
    name = name.lower()
    if "moltype" in kwargs and kwargs.get("moltype") is None:
        kwargs.pop("moltype")
    if name not in _calculators:
        raise ValueError(f'Unknown pairwise distance calculator "{name}"')

    calc = _calculators[name]
    return calc(*args, **kwargs)


def available_distances():
    """returns Table listing available fast pairwise genetic distance calculator

    Notes
    -----
    For more complicated genetic distance methods, see the evolve.models module.
    """
    from cogent3.util.table import Table

    rows = []
    for n, c in _calculators.items():
        rows.append([n, ", ".join(c.valid_moltypes)])

    table = Table(
        header=["Abbreviation", "Suitable for moltype"],
        data=rows,
        title=(
            "Specify a pairwise genetic distance calculator "
            "using 'Abbreviation' (case insensitive)."
        ),
        index_name="Abbreviation",
    )
    return table


class DistanceMatrix(DictArray):
    """pairwise distance matrix"""

    def __init__(self, dists, invalid=None):
        super(DistanceMatrix, self).__init__(dists, dtype=float)

        self._invalid = invalid

    def __setitem__(self, names, value):
        (index, remaining) = self.template.interpret_index(names)
        self.array[index] = value
        return

    def __getitem__(self, names):
        (index, remaining) = self.template.interpret_index(names)
        result = self.array[index]
        if remaining is not None:
            result = self.__class__(result, remaining)
            self.template.names = array(self.template.names)[index]
            result.template = self.template
        return result

    @property
    def names(self):
        return self.template.names[0]

    def to_table(self):
        """converted to a Table"""
        from cogent3.util.table import Table

        data = {"names": self.names}
        for i, name in enumerate(self.names):
            column = self.array[:, i]
            data[name] = column
        header = ["names"] + list(self.names)
        return Table(header=header, data=data, index_name="names")

    def to_dict(self, **kwargs):
        """Returns a flattened dict with diagonal elements removed"""
        result = super(DistanceMatrix, self).to_dict(flatten=True)
        for n1 in self.names:
            del result[(n1, n1)]
        return result

    def to_rich_dict(self):
        # because dicts with tuples as keys cannot be json'ed, we convert to
        # a list of tuples
        dists = self.to_dict()
        json_safe = [(k[0], k[1], dists[k]) for k in dists]
        return dict(
            dists=json_safe,
            invalid=self._invalid,
            type=get_object_provenance(self),
            version=__version__,
        )

    def take_dists(self, names, negate=False):
        """
        Parameters
        ----------
        names
            series of names
        negate : bool
            if True, elements in names will be excluded
        Returns
        -------
        DistanceMatrix for names x names
        """
        if type(names) == str:
            names = [names]

        current_names = array(self.names)
        if negate:
            keep = [i for i, n in enumerate(current_names) if n not in names]
        else:
            keep = [i for i, n in enumerate(current_names) if n in names]

        data = self.array.take(keep, axis=0)
        data = data.take(keep, axis=1)
        names = current_names.take(keep)
        dists = {
            (names[i], names[j]): data[i, j]
            for i in range(len(names))
            for j in range(len(names))
            if i != j
        }
        if not dists:
            result = None
        else:
            result = self.__class__(dists)
        return result

    def drop_invalid(self):
        """drops all rows / columns with an invalid entry"""
        if (
            self.shape[0] != self.shape[1]
            or self.template.names[0] != self.template.names[1]
        ):
            raise RuntimeError("Must be a square matrix")
        names = array(self.names)
        # NaN is an invalid value
        cols = numpy.isnan(self.array).sum(axis=0)
        exclude = names[cols != 0].tolist()
        rows = numpy.isnan(self.array).sum(axis=1)
        exclude += names[rows != 0].tolist()
        exclude = set(exclude)
        keep = set(names) ^ exclude
        return self.take_dists(keep)

    def quick_tree(self, show_progress=False):
        """returns a neighbour joining tree
        Returns
        -------
        an estimated Neighbour Joining Tree, note that invalid distances are dropped
        prior to building the tree
        """
        from cogent3.phylo.nj import nj

        dists = self.drop_invalid()
        if not dists or dists.shape[0] == 1:
            raise ValueError("Too few distances to build a treenj")
        dists = dists.to_dict()
        return nj(dists, show_progress=show_progress)



#!/usr/bin/env python

import warnings

import numpy

from cogent3.evolve.likelihood_tree import make_likelihood_tree_leaf
from cogent3.recalculation.definition import CalcDefn, PartitionDefn


__author__ = "Peter Maxwell"
__copyright__ = "Copyright 2007-2022, The Cogent Project"
__credits__ = ["Peter Maxwell"]
__license__ = "BSD-3"
__version__ = "2023.2.12a1"
__maintainer__ = "Gavin Huttley"
__email__ = "gavin.huttley@anu.edu.au"
__status__ = "Production"


def make_model(mprob_model, tuple_alphabet, mask):
    if mprob_model == "monomers":
        return PosnSpecificMonomerProbModel(tuple_alphabet, mask)
    elif mprob_model == "monomer":
        return MonomerProbModel(tuple_alphabet, mask)
    elif mprob_model == "conditional":
        return ConditionalMotifProbModel(tuple_alphabet, mask)
    elif mprob_model in ["word", "tuple", None]:
        return SimpleMotifProbModel(tuple_alphabet)
    else:
        raise ValueError(f"Unknown mprob model '{str(mprob_model)}'")


class MotifProbModel(object):
    def __init__(self, *whatever, **kw):
        raise NotImplementedError

    def calc_word_probs(self, *monomer_probs):
        assert len(monomer_probs) == 1
        return monomer_probs[0]

    def calc_word_weight_matrix(self, *monomer_probs):
        assert len(monomer_probs) == 1
        return monomer_probs[0]

    def make_motif_probs_defn(self):
        """Makes the first part of a parameter controller definition for this
        model, the calculation of motif probabilities"""
        return PartitionDefn(
            name="mprobs",
            default=None,
            dimensions=("locus", "edge"),
            dimension=("motif", tuple(self.get_input_alphabet())),
        )

    def set_param_controller_motif_probs(self, pc, motif_probs, **kw):
        pc.set_param_rule("mprobs", value=motif_probs, **kw)

    def count_motifs(self, alignment, include_ambiguity=False, recode_gaps=True):
        result = None
        try:
            mtype = self.alphabet.moltype
        except AttributeError:
            mtype = self.monomer_alphabet.moltype

        for seq_name in alignment.names:
            sequence = alignment.get_gapped_seq(seq_name, recode_gaps, moltype=mtype)
            leaf = make_likelihood_tree_leaf(
                sequence, self.get_counted_alphabet(), seq_name
            )
            count = leaf.get_motif_counts(include_ambiguity=include_ambiguity)
            if result is None:
                result = count.copy()
            else:
                result += count
        return result

    def adapt_motif_probs(self, motif_probs, auto=False):
        motif_probs = self.get_input_alphabet().adapt_motif_probs(motif_probs)
        assert abs(sum(motif_probs) - 1.0) < 0.0001, motif_probs
        return motif_probs

    def make_equal_motif_probs(self):
        alphabet = self.get_input_alphabet()
        p = 1.0 / len(alphabet)
        return dict([(m, p) for m in alphabet])

    def make_sample_motif_probs(self):
        import random

        motif_probs = numpy.array(
            [random.uniform(0.2, 1.0) for m in self.get_counted_alphabet()]
        )
        motif_probs /= sum(motif_probs)
        return motif_probs


class SimpleMotifProbModel(MotifProbModel):
    def __init__(self, alphabet):
        self.alphabet = alphabet

    def get_input_alphabet(self):
        return self.alphabet

    def get_counted_alphabet(self):
        return self.alphabet

    def make_motif_word_prob_defns(self):
        monomer_probs = self.make_motif_probs_defn()
        return (monomer_probs, monomer_probs, monomer_probs)


class ComplexMotifProbModel(MotifProbModel):
    def __init__(self, tuple_alphabet, mask):
        """Arguments:
        - tuple_alphabet: series of multi-letter motifs
        - monomers: the monomers from which the motifs are made
        - mask: instantaneous change matrix"""
        self.mask = mask
        self.tuple_alphabet = tuple_alphabet
        self.monomer_alphabet = monomers = tuple_alphabet.moltype.alphabet
        self.word_length = length = tuple_alphabet.get_motif_len()
        size = len(tuple_alphabet)

        # m2w[AC, 1] = C
        # w2m[0, AC, A] = True
        # w2c[ATC, AT*] = 1
        self.m2w = m2w = numpy.zeros([size, length], int)
        self.w2m = w2m = numpy.zeros([length, size, len(monomers)], int)
        contexts = monomers.get_word_alphabet(length - 1)
        self.w2c = w2c = numpy.zeros([size, length * len(contexts)], int)
        for (i, word) in enumerate(tuple_alphabet):
            for j in range(length):
                monomer = monomers.index(word[j])
                context = contexts.index(word[:j] + word[j + 1 :])
                m2w[i, j] = monomer
                w2m[j, i, monomer] = 1
                w2c[i, context * length + j] = 1

        self.mutated_posn = numpy.zeros(mask.shape, int)
        self.mutant_motif = numpy.zeros(mask.shape, int)
        self.context_indices = numpy.zeros(mask.shape, int)

        for (i, old_word, j, new_word, diff) in self._mutations():
            self.mutated_posn[i, j] = diff
            mutant_motif = new_word[diff]
            context = new_word[:diff] + new_word[diff + 1 :]
            self.mutant_motif[i, j] = monomers.index(mutant_motif)
            c = contexts.index(context)
            self.context_indices[i, j] = c * length + diff

    def _mutations(self):
        diff_pos = lambda x, y: [i for i in range(len(x)) if x[i] != y[i]]
        num_states = len(self.tuple_alphabet)
        for i in range(num_states):
            old_word = self.tuple_alphabet[i]
            for j in range(num_states):
                new_word = self.tuple_alphabet[j]
                if self.mask[i, j]:
                    assert self.mask[i, j] == 1.0
                    diffs = diff_pos(old_word, new_word)
                    assert len(diffs) == 1, (old_word, new_word)
                    diff = diffs[0]
                    yield i, old_word, j, new_word, diff


class MonomerProbModel(ComplexMotifProbModel):
    def get_input_alphabet(self):
        return self.monomer_alphabet

    def get_counted_alphabet(self):
        return self.monomer_alphabet

    def calc_monomer_probs(self, word_probs):
        monomer_probs = numpy.dot(word_probs, self.w2m.sum(axis=0))
        monomer_probs /= monomer_probs.sum()
        return monomer_probs

    def calc_word_probs(self, monomer_probs):
        result = numpy.product(monomer_probs.take(self.m2w), axis=-1)
        # maybe simpler but slower, works ok:
        # result = numpy.product(monomer_probs ** (w2m, axis=-1))
        result /= result.sum()
        return result

    def calc_word_weight_matrix(self, monomer_probs):
        return monomer_probs.take(self.mutant_motif) * self.mask

    def make_motif_word_prob_defns(self):
        monomer_probs = self.make_motif_probs_defn()
        word_probs = CalcDefn(self.calc_word_probs, name="wprobs")(monomer_probs)
        mprobs_matrix = CalcDefn(self.calc_word_weight_matrix, name="mprobs_matrix")(
            monomer_probs
        )
        return (monomer_probs, word_probs, mprobs_matrix)

    def adapt_motif_probs(self, motif_probs, auto=False):
        try:
            motif_probs = self.monomer_alphabet.adapt_motif_probs(motif_probs)
        except ValueError:
            motif_probs = self.tuple_alphabet.adapt_motif_probs(motif_probs)
            if not auto:
                warnings.warn("Motif probs overspecified", stacklevel=5)
            motif_probs = self.calc_monomer_probs(motif_probs)
        return motif_probs


class PosnSpecificMonomerProbModel(MonomerProbModel):
    def get_counted_alphabet(self):
        return self.tuple_alphabet

    def calc_posn_specific_monomer_probs(self, word_probs):
        monomer_probs = numpy.dot(word_probs, self.w2m)
        monomer_probs /= monomer_probs.sum(axis=1)[..., numpy.newaxis]
        return list(monomer_probs)

    def calc_word_probs(self, monomer_probs):
        positions = list(range(self.word_length))
        assert len(monomer_probs) == self.m2w.shape[1], (
            len(monomer_probs),
            type(monomer_probs),
            self.m2w.shape,
        )
        result = numpy.product(
            [monomer_probs[i].take(self.m2w[:, i]) for i in positions], axis=0
        )
        result /= result.sum()
        return result

    def calc_word_weight_matrix(self, monomer_probs):
        monomer_probs = numpy.array(monomer_probs)  # so [posn, motif]
        size = monomer_probs.shape[-1]
        # should be constant
        extended_indices = self.mutated_posn * size + self.mutant_motif
        result = monomer_probs.take(extended_indices) * self.mask
        return result

    def make_motif_word_prob_defns(self):
        monomer_probs = PartitionDefn(
            name="psmprobs",
            default=None,
            dimensions=("locus", "position", "edge"),
            dimension=("motif", tuple(self.get_input_alphabet())),
        )
        monomer_probs3 = monomer_probs.across_dimension(
            "position", [str(i) for i in range(self.word_length)]
        )
        monomer_probs3 = CalcDefn(lambda *x: numpy.array(x), name="mprobs")(
            *monomer_probs3
        )
        word_probs = CalcDefn(self.calc_word_probs, name="wprobs")(monomer_probs3)
        mprobs_matrix = CalcDefn(self.calc_word_weight_matrix, name="mprobs_matrix")(
            monomer_probs3
        )
        return (monomer_probs, word_probs, mprobs_matrix)

    def set_param_controller_motif_probs(self, pc, motif_probs, **kw):
        assert len(motif_probs) == self.word_length
        for (i, m) in enumerate(motif_probs):
            pc.set_param_rule("psmprobs", value=m, position=str(i), **kw)

    def adapt_motif_probs(self, motif_probs, auto=False):
        try:
            motif_probs = self.monomer_alphabet.adapt_motif_probs(motif_probs)
        except ValueError:
            motif_probs = self.tuple_alphabet.adapt_motif_probs(motif_probs)
            motif_probs = self.calc_posn_specific_monomer_probs(motif_probs)
        else:
            motif_probs = [motif_probs] * self.word_length
        return motif_probs


class ConditionalMotifProbModel(ComplexMotifProbModel):
    def get_input_alphabet(self):
        return self.tuple_alphabet

    def get_counted_alphabet(self):
        return self.tuple_alphabet

    def calc_word_weight_matrix(self, motif_probs):
        context_probs = numpy.dot(motif_probs, self.w2c)
        context_probs[context_probs == 0.0] = numpy.inf
        return motif_probs / context_probs.take(self.context_indices)

    def make_motif_word_prob_defns(self):
        mprobs = self.make_motif_probs_defn()
        mprobs_matrix = CalcDefn(self.calc_word_weight_matrix, name="mprobs_matrix")(
            mprobs
        )
        return (mprobs, mprobs, mprobs_matrix)



from numpy import array

from cogent3.core.moltype import get_moltype
from cogent3.core.profile import MotifFreqsArray


__author__ = "Gavin Huttley"
__copyright__ = "Copyright 2007-2012, The Cogent Project"
__credits__ = ["Gavin Huttley"]
__license__ = "BSD-3"
__version__ = "2023.2.12a1"
__maintainer__ = "Gavin Huttley"
__email__ = "gavin.huttley@anu.edu.au"
__status__ = "Alpha"


def read(filepath):
    """returns MotifFreqsArray matrix"""
    try:
        infile = open(filepath)
        data = infile.readlines()
        infile.close()
    except TypeError:
        data = filepath

    data = [l.split() for l in data]
    revised = list(zip(*data))
    states = []
    matrix = []
    for row in revised[1:]:
        states.append(row[0])
        matrix.append([float(i) for i in row[1:]])

    matrix = dict(zip(states, matrix))
    if len(states) == 4:
        name = "rna" if "U" in states else "dna"
    else:
        name = "protein"

    states = list(get_moltype(name))
    matrix = [matrix[s] for s in states]
    matrix = array(matrix, dtype=float)

    pfm = MotifFreqsArray(matrix.T, states)
    return pfm



#!/usr/bin/env python
"""Provides support functions and classes for parsers.
"""
from copy import deepcopy

from cogent3.util.misc import iterable


__author__ = "Rob Knight"
__copyright__ = "Copyright 2007-2022, The Cogent Project"
__credits__ = ["Rob Knight", "Peter Maxwell"]
__license__ = "BSD-3"
__version__ = "2023.2.12a1"
__maintainer__ = "Gavin Huttley"
__email__ = "Gavin.Huttley@anu.edu.au"
__status__ = "Development"


class FileFormatError(Exception):
    """Exception raised when a file can not be parsed."""

    pass


class RecordError(FileFormatError):
    """Exception raised when a record is bad."""

    pass


class FieldError(RecordError):
    """Exception raised when a field within a record is bad."""

    pass


class Grouper(object):
    """Acts as iterator that returns lists of n items at a time from seq.

    Note: returns a partial list if not evenly divisible by n.
    """

    def __init__(self, NumItems=1):
        """Returns new Grouper object: will return n items at a time from seq"""
        self.NumItems = NumItems

    def __call__(self, seq):
        """Returns iterator over seq, returning n items at a time."""
        try:
            num = int(self.NumItems)
            assert num >= 1
        except:
            raise ValueError(
                f"Grouper.NumItems must be positive int, not {self.NumItems}"
            )
        curr = []
        for i, item in enumerate(seq):
            if (i % num == 0) and curr:
                yield curr
                curr = [item]
            else:
                curr.append(item)
        # return any leftover items
        if curr:
            yield curr


# Example of the instances Grouper provides:
ByPairs = Grouper(2)


def string_and_strip(*items):
    """Converts items to strings and strips them."""
    return [str(i).strip() for i in items]


def DelimitedSplitter(delimiter=None, max_splits=1):
    """Returns function that returns stripped fields split by delimiter.

    Unlike the default behavior of split, max_splits can be negative, in
    which case it counts from the end instead of the start (i.e. splits
    at the _last_ delimiter, last two delimiters, etc. for -1, -2, etc.)
    However, if the delimiter is None (the default) and max_splits is
    negative, will not preserve internal spaces.

    Note: leaves empty fields in place.
    """
    is_int = isinstance(max_splits, int) or isinstance(max_splits, int)
    if is_int and (max_splits > 0):

        def parser(line):
            return [i.strip() for i in line.split(delimiter, max_splits)]

    elif is_int and (max_splits < 0):

        def parser(line):
            to_insert = delimiter or " "  # re-join fields w/ space if None
            fields = line.split(delimiter)
            if (fields == []) or (fields == [""]):
                return []  # empty string or only delimiter: return nothing
            # if not enough fields, count from the start, not the end
            if len(fields) < max_splits:
                first_fields = fields[0]
                last_fields = fields[1:]
            # otherwise, count off the last n fields and join the remainder
            else:
                first_fields = fields[:max_splits]
                last_fields = fields[max_splits:]
            pieces = []
            # if first_fields is empty, don't make up an extra empty string
            if first_fields:
                pieces.append(to_insert.join(first_fields))
            pieces.extend(last_fields)
            return [i.strip() for i in pieces]

    else:  # ignore max_splits if it was 0

        def parser(line):
            return [i.strip() for i in line.split(delimiter)]

    return parser


# The following provide examples of the kinds of functions DelimitedSplitter
# returns.
semi_splitter = DelimitedSplitter(";", None)
space_pairs = DelimitedSplitter(None)
equal_pairs = DelimitedSplitter("=")
last_colon = DelimitedSplitter(":", -1)


class GenericRecord(dict):
    """Holds data for a generic field ->: value mapping.

    Override Required with {name:prototype} mapping. Each required name
    will get a deepcopy of its prototype. For example, use an empty list to
    guarantee that each instance has its own list for a particular field to
    which items can be appended.

    Raises AttributeError on attempt to delete required item, but does not
    raise an exception on attempt to delete absent item.

    This class explicitly does _not_ override __getitem__ or __setitem__ for
    performance reasons: if you need to transform keys on get/set or if you
    need to access items as attributes and vice versa, use MappedRecord
    instead.
    """

    Required = {}

    def __init__(self, *args, **kwargs):
        """Reads kwargs as properties of self."""
        # perform init on temp dict to preserve interface: will then translate
        # aliased keys when loading into self
        temp = {}
        dict.__init__(temp, *args, **kwargs)
        self.update(temp)
        for name, prototype in self.Required.items():
            if name not in self:
                self[name] = deepcopy(prototype)

    def __delitem__(self, item):
        """Deletes item or raises exception if item required.

        Note: Fails silently if item absent.
        """
        if item in self.Required:
            raise AttributeError(f"{item} is a required item")
        try:
            super(GenericRecord, self).__delitem__(item)
        except KeyError:
            pass

    def copy(self):
        """Coerces copy to correct type"""
        temp = self.__class__(super(GenericRecord, self).copy())
        # don't forget to copy attributes!
        for attr, val in self.__dict__.items():
            temp.__dict__[attr] = deepcopy(val)
        return temp


class MappedRecord(GenericRecord):
    """GenericRecord that maps names of fields onto standardized names.

    Override Aliases in subclass for new mapping of OldName->NewName. Each
    OldName can have only one NewName, but it's OK if several OldNames map
    to the same NewName.

    Note: can access fields either as items or as attributes. In addition,
    can access either using nonstandard names or using standard names.

    Implementation note: currently, just a dict with appropriate get/set
    overrides and ability to access items as attributes. Attribute access
    is about 10x slower than in GenericRecord, so make sure you need the
    additional capabilities if you use MappedRecord instead of GenericRecord.

    WARNING: MappedRecord pretends to have every attribute, so will never raise
    AttributeError when trying to find an unknown attribute. This feature can
    cause surprising interactions when a Delegator is delegating its
    attributes to a MappedRecord, since any attributes defined in __init__ will
    be set in the MappedRecord and not in the object itself. The solution is
    to use the self.__dict__['AttributeName'] = foo syntax to force the
    attributes to be set in the object and not the MappedRecord to which it
    forwards.
    """

    Aliases = {}

    DefaultValue = None

    def _copy(self, prototype):
        """Returns a copy of item."""
        if hasattr(prototype, "copy"):
            return prototype.copy()
        elif isinstance(prototype, list):
            return prototype[:]
        elif (
            isinstance(prototype, str)
            or isinstance(prototype, int)
            or isinstance(prototype, int)
            or isinstance(prototype, tuple)
            or isinstance(prototype, complex)
            or prototype is None
        ):
            return prototype  # immutable type: use directly
        else:
            return deepcopy(prototype)

    def __init__(self, *args, **kwargs):
        """Reads kwargs as properties of self."""
        # perform init on temp dict to preserve interface: will then translate
        # aliased keys when loading into self
        temp = {}
        unalias = self.unalias
        dict.__init__(temp, *args, **kwargs)
        for key, val in temp.items():
            self[unalias(key)] = val
        for name, prototype in self.Required.items():
            new_name = unalias(name)
            if new_name not in self:
                self[new_name] = self._copy(prototype)

    def unalias(self, key):
        """Returns dealiased name for key, or key if not in alias."""
        try:
            return self.Aliases.get(key, key)
        except TypeError:
            return key

    def __getattr__(self, attr):
        """Returns None if field is absent, rather than raising exception."""
        if attr in self:
            return self[attr]
        elif attr in self.__dict__:
            return self.__dict__[attr]
        elif attr.startswith("__"):  # don't retrieve private class attrs
            raise AttributeError
        elif hasattr(self.__class__, attr):
            return getattr(self.__class__, attr)
        else:
            return self._copy(self.DefaultValue)

    def __setattr__(self, attr, value):
        """Sets attribute in self if absent, converting name if necessary."""
        normal_attr = self.unalias(attr)
        # we overrode __getattr__, so have to simulate getattr(self, attr) by
        # calling superclass method and checking for AttributeError.
        # BEWARE: dict defines __getattribute__, not __getattr__!
        try:
            super(MappedRecord, self).__getattribute__(normal_attr)
            super(MappedRecord, self).__setattr__(normal_attr, value)
        except AttributeError:
            self[normal_attr] = value

    def __delattr__(self, attr):
        """Deletes attribute, converting name if necessary. Fails silently."""
        normal_attr = self.unalias(attr)
        if normal_attr in self.Required:
            raise AttributeError(f"{attr} is a required attribute")
        else:
            try:
                super(MappedRecord, self).__delattr__(normal_attr)
            except AttributeError:
                del self[normal_attr]

    def __getitem__(self, item):
        """Returns default if item is absent, rather than raising exception."""
        normal_item = self.unalias(item)
        return self.get(normal_item, self._copy(self.DefaultValue))

    def __setitem__(self, item, val):
        """Sets item, converting name if necessary."""
        super(MappedRecord, self).__setitem__(self.unalias(item), val)

    def __delitem__(self, item):
        """Deletes item, converting name if necessary. Fails silently."""
        normal_item = self.unalias(item)
        super(MappedRecord, self).__delitem__(normal_item)

    def __contains__(self, item):
        """Tests membership, converting name if necessary."""
        return super(MappedRecord, self).__contains__(self.unalias(item))

    def get(self, item, default):
        """Returns self[item] or default if not present. Silent when unhashable."""
        try:
            return super(MappedRecord, self).get(self.unalias(item), default)
        except TypeError:
            return default

    def setdefault(self, key, default=None):
        """Returns self[key] or default (and sets self[key]=default)"""
        return super(MappedRecord, self).setdefault(self.unalias(key), default)

    def update(self, *args, **kwargs):
        """Updates self with items in other"""
        temp = {}
        unalias = self.unalias
        temp.update(*args, **kwargs)
        for key, val in temp.items():
            self[unalias(key)] = val


# The following methods are useful for handling particular types of fields in
# line-oriented parsers


def TypeSetter(constructor=None):
    """Returns function that takes obj, field, val and sets obj.field = val.

    constructor can be any callable that returns an object.
    """
    if constructor:

        def setter(obj, field, val):
            setattr(obj, field, constructor(val))

    else:

        def setter(obj, field, val):
            setattr(obj, field, val)

    return setter


int_setter = TypeSetter(int)
str_setter = TypeSetter(str)
list_setter = TypeSetter(list)
tuple_setter = TypeSetter(tuple)
dict_setter = TypeSetter(dict)
float_setter = TypeSetter(float)
complex_setter = TypeSetter(complex)
bool_setter = TypeSetter(bool)
identity_setter = TypeSetter()


def list_adder(obj, field, val):
    """Adds val to list in obj.field, creating list if necessary."""
    try:
        getattr(obj, field).append(val)
    except AttributeError:
        setattr(obj, field, [val])


def list_extender(obj, field, val):
    """Adds val to list in obj.field, creating list if necessary."""
    try:
        getattr(obj, field).extend(iterable(val))
    except AttributeError:
        setattr(obj, field, list(val))


def dict_adder(obj, field, val):
    """If val is a sequence, adds key/value pair in obj.field: else adds key."""
    try:
        key, value = val
    except (ValueError, TypeError):
        key, value = val, None
    try:
        getattr(obj, field)[key] = value
    except AttributeError:
        setattr(obj, field, {key: value})


class LineOrientedConstructor(object):
    """Constructs a MappedRecord from a sequence of lines."""

    def __init__(
        self,
        Lines=None,
        LabelSplitter=space_pairs,
        FieldMap=None,
        Constructor=MappedRecord,
        Strict=False,
    ):
        """Returns new LineOrientedConstructor.

        Fields:
            Lines: set of lines to construct record from (for convenience).
            Default is None.

            LabelSplitter: function that returns (label, data) tuple.
            Default is to split on first space and strip components.

            FieldMap: dict of {fieldname:handler} functions. Each function
            has the signature (obj, field, val) and performs an inplace
            action like setting field to val or appending val to field.
            Default is empty dict.

            Constructor: constructor for the resulting object.
            Default is MappedRecord: beware of using constructors that don't
            subclass MappedRecord.

            Strict: boolean controlling whether to raise error on unrecognized
            field. Default is False.
        """
        self.Lines = Lines or []
        self.LabelSplitter = LabelSplitter
        self.FieldMap = FieldMap or {}
        self.Constructor = Constructor
        self.Strict = Strict

    def __call__(self, Lines=None):
        """Returns the record constructed from Lines, or self.Lines"""
        if Lines is None:
            Lines = self.Lines
        result = self.Constructor()
        fieldmap = self.FieldMap
        aka = result.unalias

        splitter = self.LabelSplitter
        for line in Lines:
            # find out how many items we got, setting key and val appropiately
            items = list(splitter(line))
            num_items = len(items)
            if num_items == 2:  # typical case: key-value pair
                raw_field, val = items
            elif num_items > 2:
                raw_field = items[0]
                val = items[1:]
            elif len(items) == 1:
                raw_field, val = items[0], None
            elif not items:  # presumably had line with just a delimiter?
                continue
            # figure out if we know the field under its original name or as
            # an alias
            if raw_field in fieldmap:
                field, mapper = raw_field, fieldmap[raw_field]
            else:
                new_field = aka(raw_field)
                if new_field in fieldmap:
                    field, mapper = new_field, fieldmap[new_field]
                else:
                    if self.Strict:
                        raise FieldError(f"Got unrecognized field {raw_field}")
                    else:
                        identity_setter(result, raw_field, val)
                    continue
            # if we found the field in the fieldmap, apply the correct function
            try:
                mapper(result, field, val)
            except:  # Warning: this is a catchall for _any_ exception,
                # and may mask what's actually going wrong.
                if self.Strict:
                    raise FieldError(f"Could not handle line {line}")
        return result


def FieldWrapper(fields, splitter=None, constructor=None):
    """Returns dict containing field->val mapping, one level.

    fields should be list of fields, in order.
    splitter should be something like a DelimitedSplitter that converts the
        line into a sequence of fields.
    constructor is a callable applied to the dict after construction.

    Call result on a _single_ line, not a list of lines.

    Note that the constructor should take a dict and return an object of some
    useful type. Additionally, it is the _constructor's_ responsibility to
    complain if there are not enough fields, since zip will silently truncate
    at the shorter sequence. This is actually useful in the case where many of
    the later fields are optional.
    """
    if splitter is None:
        splitter = DelimitedSplitter(None, None)
    if constructor:

        def parser(line):
            return constructor(dict(list(zip(fields, splitter(line)))))

    else:

        def parser(line):
            return dict(list(zip(fields, splitter(line))))

    return parser


def StrictFieldWrapper(fields, splitter=None, constructor=None):
    """Returns dict containing field->val mapping, one level.

    fields should be list of fields, in order.
    splitter should be something like a DelimitedSplitter that converts the
        line into a sequence of fields.
    constructor is a callable applied to the dict after construction.

    Call result on a _single_ line, not a list of lines.

    Note that the constructor should take a dict and return an object of some
    useful type. Raises RecordError if the wrong number of fields are returned
    from the split.
    """
    if splitter is None:
        splitter = DelimitedSplitter(None, None)
    if constructor:

        def parser(line):
            items = splitter(line)
            if len(items) != len(fields):
                raise FieldError(
                    f"Expected {len(fields)} items but got {len(items)}: {items}"
                )
            return constructor(dict(list(zip(fields, items))))

    else:

        def parser(line):
            items = splitter(line)
            if len(items) != len(fields):
                raise FieldError(
                    f"Expected {len(fields)} items but got {len(items)}: {items}"
                )
            return dict(list(zip(fields, items)))

    return parser


def raise_unknown_field(field, data):
    """Raises a FieldError, displaying the offending field and data."""
    raise FieldError(f"Got unknown field {field} with data {data}")


class FieldMorpher(object):
    """When called, applies appropriate constructors to each value of dict.

    Initialize using a dict of fieldname:constructor pairs.
    """

    def __init__(self, Constructors, Default=raise_unknown_field):
        """Returns a new FieldMorpher, using appropriate constructors.

        If a field is unknown, will try to set key and value to the results
        of Default(key, value): in other words, the signature of Default should
        take a key and a value and should return a key and a value. The
        built-in value of Default raises a FieldError instead, but it will
        often be useful to do things like return the key/value pair unchanged,
        or to strip the key and the value and then add them.
        """
        self.Constructors = Constructors
        self.Default = Default

    def __call__(self, data):
        """Returns a new dict containing information converted from data."""
        result = {}
        default = self.Default
        cons = self.Constructors
        for key, val in data.items():
            if key in cons:
                result[key] = cons[key](val)
            else:
                new_key, new_val = default(key, val)
                # if we now recognize the key, use its constructor on the old
                # val
                if new_key in cons:
                    result[new_key] = cons[new_key](val)
                # otherwise, enter the new key and the new val
                else:
                    result[new_key] = new_val
        return result
